import json
import os
import subprocess
import time

from datetime import datetime, timezone
from threading import Thread

import pytest
import requests

import runhouse as rh

from runhouse.constants import (
    DEFAULT_HTTP_PORT,
    DEFAULT_HTTPS_PORT,
    DEFAULT_PROCESS_NAME,
    DEFAULT_SERVER_PORT,
    HOUR,
    LAST_ACTIVE_AT_TIMEFRAME,
    LOCALHOST,
    MINUTE,
)

from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import ClusterStatus, RunhouseDaemonStatus

from runhouse.resources.images.image import ImageSetupStepType

from runhouse.utils import _process_env_vars, get_random_str

import tests.test_resources.test_resource
from tests.conftest import init_args
from tests.fixtures.resource_fixtures import create_folder_path
from tests.utils import (
    _get_env_var_value,
    friend_account,
    friend_account_in_org,
    keep_config_keys,
    set_daemon_and_cluster_status,
    set_output_env_vars,
)

""" TODO:
1) In subclasses, test factory methods create same type as parent
2) In subclasses, use monkeypatching to make sure `up()` is called for various methods if the server is not up
3) Test AWS, GCP, and Azure static clusters separately
"""


def load_shared_resource_config(resource_class_name, address):
    resource_class = getattr(rh, resource_class_name)
    loaded_resource = resource_class.from_name(address, dryrun=True)
    return loaded_resource.config()


def summer(a: int, b: int):
    return a + b


def sub(a: int, b: int):
    return a - b


def cluster_keys(cluster):
    return cluster.keys()


def cluster_config():
    return rh.here.config()


def assume_caller_and_get_token():
    token_default = rh.configs.token
    with rh.as_caller():
        token_as_caller = rh.configs.token
    return token_default, token_as_caller


def sleep_fn(secs):
    import time

    time.sleep(secs)


def import_env():
    import pandas  # noqa
    import pytest  # noqa

    return "success"


def run_in_no_env(cmd):
    return rh.here.run_bash(cmd)


def run_node_all(cmd):
    # This forces `cluster.run` to use ssh instead of calling an env run
    return rh.here.run_bash(cmd, node="all")


def sort_env_servlet_processes(env_servlet_processes: dict):
    """helping function for the test_send_status_to_db test, sort the servlet_processed dict (including its sub
    dicts) by their keys."""
    keys = list(env_servlet_processes.keys())
    keys.sort()
    sorted_env_servlet_processes = {}
    for k in keys:
        sub_keys = list(env_servlet_processes[k].keys())
        sub_keys.sort()
        nested_dict = {i: env_servlet_processes[k][i] for i in sub_keys}
        sorted_env_servlet_processes[k] = nested_dict
    return sorted_env_servlet_processes


class TestCluster(tests.test_resources.test_resource.TestResource):
    MAP_FIXTURES = {"resource": "cluster"}

    UNIT = {"cluster": ["named_cluster"]}
    LOCAL = {
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",  # Represents private dev use case
            "docker_cluster_pk_ssh_den_auth",  # Helps isolate Auth issues
            "docker_cluster_pk_http_exposed",  # Represents within VPC use case
            "docker_cluster_pwd_ssh_no_auth",
        ],
    }
    MINIMAL = {"cluster": ["static_cpu_pwd_cluster"]}
    RELEASE = {
        "cluster": ["static_cpu_pwd_cluster", "static_cpu_pwd_cluster_den_launcher"]
    }
    MAXIMAL = {
        "cluster": [
            "docker_cluster_pk_ssh_no_auth",
            "docker_cluster_pk_ssh_den_auth",
            "docker_cluster_pwd_ssh_no_auth",
            "static_cpu_pwd_cluster",
            "multinode_cpu_docker_conda_cluster",
            "static_gpu_pwd_cluster_den_launcher",
        ]
    }

    @pytest.mark.level("unit")
    @pytest.mark.clustertest
    def test_cluster_factory_and_properties(self, cluster):
        assert isinstance(cluster, rh.Cluster)
        args = init_args[id(cluster)]
        if "ips" in args:
            # Check that it's a Cluster and not a subclass
            assert cluster.__class__.name == "Cluster"
            assert cluster.ips == args["ips"]
            assert cluster.head_ip == args["ips"][0]

        if "ssh_creds" in args:
            args_creds = args["ssh_creds"]
            args_creds_values = (
                args_creds.values if isinstance(args_creds, rh.Secret) else args_creds
            )

            cluster_creds = cluster.creds_values
            if "ssh_private_key" in cluster_creds:
                # this means that the secret was created by accessing an ssh-key file
                cluster_creds.pop("private_key", None)
                cluster_creds.pop("public_key", None)
            assert cluster_creds == args_creds_values

        if "server_host" in args:
            assert cluster.server_host == args["server_host"]
        else:
            assert cluster.server_host is None

        if "ssl_keyfile" in args:
            assert cluster.cert_config.key_path == args["ssl_keyfile"]

        if "ssl_certfile" in args:
            assert cluster.cert_config.cert_path == args["ssl_certfile"]

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_recreate(self, cluster):
        # # Create underlying ssh connection if not already
        if cluster.ips == cluster.internal_ips != ["localhost"]:
            cluster.run_bash_over_ssh(["echo hello"])
        else:
            cluster.run_bash(["echo hello"])

        num_open_tunnels = len(rh.globals.ssh_tunnel_cache)

        # Create a new cluster object for the same remote cluster
        cluster.save()
        new_cluster = rh.cluster(cluster.rns_address)
        if new_cluster.ips == new_cluster.internal_ips != ["localhost"]:
            new_cluster.run_bash_over_ssh(["echo hello"])
        else:
            new_cluster.run_bash(["echo hello"])

        # Check that the same underlying ssh connection was used
        assert len(rh.globals.ssh_tunnel_cache) == num_open_tunnels

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_endpoint(self, cluster):
        if not cluster.ips:
            assert cluster.endpoint() is None
            return

        endpoint = cluster.endpoint()
        if cluster.server_connection_type == "ssh":
            assert cluster.endpoint(external=True) is None
            assert endpoint == f"http://{LOCALHOST}:{cluster.client_port}"
        else:
            url_base = "https" if cluster.server_connection_type == "tls" else "http"
            if cluster.client_port not in [DEFAULT_HTTP_PORT, DEFAULT_HTTPS_PORT]:
                assert (
                    endpoint
                    == f"{url_base}://{cluster.server_address}:{cluster.client_port}"
                )
            else:
                assert endpoint == f"{url_base}://{cluster.server_address}"

        # Try to curl docs
        verify = cluster.client.verify
        r = requests.get(
            f"{endpoint}/status",
            verify=verify,
            headers=rh.globals.rns_client.request_headers(),
        )
        assert r.status_code == 200
        status_data = r.json()[
            0
        ]  # getting the first element because the endpoint returns the status + response to den.
        assert status_data.get("cluster_config").get(
            "resource_subtype"
        ) == cluster.config().get("resource_subtype")
        assert status_data.get("env_servlet_processes", None)
        assert isinstance(status_data.get("server_cpu_utilization", None), float)
        assert status_data.get("server_memory_usage", None)
        assert not status_data.get("server_gpu_usage", None)

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_request_timeout(self, cluster):
        with pytest.raises(requests.exceptions.ReadTimeout):
            cluster._http_client.request_json(
                endpoint="/status",
                req_type="get",
                timeout=0.005,
                headers=rh.globals.rns_client.request_headers(),
            )

        status_res = cluster._http_client.request_json(
            endpoint="/status",
            req_type="get",
            headers=rh.globals.rns_client.request_headers(),
        )
        assert status_res

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_objects(self, cluster):
        k1 = get_random_str()
        k2 = get_random_str()
        cluster.put(k1, "v1")
        cluster.put(k2, "v2")
        assert k1 in cluster.keys()
        assert k2 in cluster.keys()
        assert cluster.get(k1) == "v1"
        assert cluster.get(k2) == "v2"

        # Make new env
        cluster.ensure_process_created("numpy_env")

        k3 = get_random_str()
        cluster.put(k3, "v3", process="numpy_env")
        assert k3 in cluster.keys()
        assert cluster.get(k3) == "v3"

    @pytest.mark.skip("pending cluster.kill functionality")
    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_delete_process(self, cluster):
        proc1 = cluster.ensure_process_created("p1")
        proc2 = cluster.ensure_process_created("p2")

        cluster.put("k1", "v1", process=proc1)
        cluster.put("k2", "v2", process=proc2)

        # test delete env2
        assert cluster.get(proc2)
        assert cluster.get("k2")

        cluster.delete(proc2)
        assert not cluster.get(proc2)
        assert not cluster.get("k2")

        assert cluster.get(proc1)
        assert cluster.get("k1")

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_condensed_config_for_cluster(self, cluster):
        remote_cluster_config = rh.function(cluster_config).to(cluster)
        on_cluster_config = remote_cluster_config()
        local_cluster_config = cluster.config()

        keys_to_keep = [
            "name",
            "is_gpu",
            "resource_subtype",
            "den_auth",
            "server_port",
            "server_connection_type",
            "autostop_mins",
            "domain",
        ]
        if local_cluster_config.get("resource_subtype", None) == "OnDemandCluster":
            keys_to_keep.append("compute_properties")
        else:
            keys_to_keep.append("ips")
        on_cluster_config = keep_config_keys(on_cluster_config, keys_to_keep)
        local_cluster_config = keep_config_keys(local_cluster_config, keys_to_keep)

        assert on_cluster_config == local_cluster_config

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_sharing(self, cluster, friend_account_logged_in_docker_cluster_pk_ssh):
        # Skip this test for ondemand clusters, because making
        # it compatible with ondemand_cluster requires changes
        # that break CI.
        # TODO: Remove this by doing some CI-specific logic.
        if cluster.__class__.__name__ == "OnDemandCluster":
            return

        if cluster.rns_address.startswith("~"):
            # For `local_named_resource` resolve the rns address so it can be shared and loaded
            from runhouse.globals import rns_client

            cluster.rns_address = rns_client.local_to_remote_address(
                cluster.rns_address
            )

        cluster.share(
            users=["info@run.house"],
            access_level="read",
            notify_users=False,
        )

        # First try loading in same process/filesystem because it's more debuggable, but not as thorough
        resource_class_name = cluster.config().get("resource_type").capitalize()
        config = cluster.config()

        sky_secret = "ssh-sky-key"
        generated_secret = f'{config["name"]}-ssh-secret'
        # CI testing can be flakey depending on den_tester user creds and overwriting: test against both values
        # expected_creds = (
        #     sky_secret
        #     if isinstance(cluster._creds, SSHSecret)
        #     else generated_secret
        # )

        with friend_account():
            curr_config = load_shared_resource_config(
                resource_class_name, cluster.rns_address
            )
            new_creds = curr_config.get("creds", None)
            assert sky_secret in new_creds or generated_secret in new_creds
            assert curr_config == config

        # TODO: If we are testing with an ondemand_cluster we to
        # sync sky key so loading ondemand_cluster from config works
        # Also need aws secret to load availability zones
        # secrets=["sky", "aws"],
        load_shared_resource_config_cluster = rh.function(
            load_shared_resource_config
        ).to(friend_account_logged_in_docker_cluster_pk_ssh)
        new_config = load_shared_resource_config_cluster(
            resource_class_name, cluster.rns_address
        )
        new_creds = curr_config.get("creds", None)
        assert sky_secret in new_creds or generated_secret in new_creds
        assert new_config == config

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_access_to_shared_cluster(self, cluster):
        # TODO: Remove this by doing some CI-specific logic.
        if cluster.__class__.__name__ == "OnDemandCluster":
            return

        if cluster.rns_address.startswith("~"):
            # For `local_named_resource` resolve the rns address so it can be shared and loaded
            from runhouse.globals import rns_client

            cluster.rns_address = rns_client.local_to_remote_address(
                cluster.rns_address
            )

        cluster.share(
            users=["support@run.house"],
            access_level="read",
            notify_users=False,
        )

        cluster_name = cluster.rns_address
        cluster_creds = cluster.creds_values

        with friend_account_in_org():
            shared_cluster = rh.cluster(name=cluster_name)
            assert shared_cluster.rns_address == cluster_name
            assert shared_cluster.creds_values.keys() == cluster_creds.keys()
            echo_msg = "hello from shared cluster"

            if shared_cluster.ips == shared_cluster.internal_ips != ["localhost"]:
                run_res = shared_cluster.run_bash_over_ssh([f"echo {echo_msg}"])
            else:
                run_res = shared_cluster.run_bash([f"echo {echo_msg}"])

            assert echo_msg in run_res[0][1]
            # First element, return code
            if shared_cluster.ips == shared_cluster.internal_ips != ["localhost"]:
                assert shared_cluster.run_bash_over_ssh(["echo hello"])[0][0] == 0
            else:
                assert shared_cluster.run_bash(["echo hello"])[0][0] == 0

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_changing_name_and_saving_in_between(self, cluster):
        remote_summer = rh.function(summer).to(cluster)
        assert remote_summer(3, 4) == 7
        old_name = cluster.name

        cluster.save(name="new_testing_name")

        assert remote_summer(3, 4) == 7
        remote_sub = rh.function(sub).to(cluster)
        assert remote_sub(3, 4) == -1

        cluster_keys_remote = rh.function(cluster_keys).to(cluster)

        # If save did not update the name, this will attempt to create a connection
        # when the cluster is used remotely. However, if you update the name, `on_this_cluster` will
        # work correctly and then the remote function will just call the object store when it calls .keys()
        assert cluster.keys() == cluster_keys_remote(cluster)

        # Restore the state?
        cluster.save(name=old_name)

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_caller_token_propagated(self, cluster):
        remote_assume_caller_and_get_token = rh.function(
            assume_caller_and_get_token
        ).to(cluster)

        remote_assume_caller_and_get_token.share(
            users=["info@run.house"], notify_users=False
        )
        with friend_account():
            unassumed_token, assumed_token = remote_assume_caller_and_get_token()
            # "Local token" is the token the cluster accesses in rh.configs.token; this is what will be used
            # in subsequent rns_client calls
            assert unassumed_token != rh.configs.token

            # Both tokens should be valid for the cluster
            assert rh.globals.rns_client.validate_cluster_token(
                assumed_token, cluster.rns_address
            )
            assert rh.globals.rns_client.validate_cluster_token(
                unassumed_token, cluster.rns_address
            )

        # Docker clusters are logged out, ondemand clusters are logged in
        if cluster.ips == cluster.internal_ips != ["localhost"]:
            output = cluster.run_bash_over_ssh(
                "sed -n 's/.*token: *//p' ~/.rh/config.yaml"
            )
        else:
            output = cluster.run_bash("sed -n 's/.*token: *//p' ~/.rh/config.yaml")
        # No config file
        if output[0] == 2:
            assert unassumed_token is None
        elif output[0] == 0:
            assert unassumed_token == output[1].strip()

    ####################################################################################################
    # Status tests
    ####################################################################################################

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_rh_status_pythonic(self, cluster):
        process = cluster.ensure_process_created("worker")
        sleep_remote = rh.function(sleep_fn).to(cluster, process=process)
        cluster.put(key="status_key1", obj="status_value1", process=process)
        # Run these in a separate thread so that the main thread can continue
        call_threads = [Thread(target=sleep_remote, args=[3]) for _ in range(3)]
        for call_thread in call_threads:
            call_thread.start()

        # Wait a second so the calls can start
        time.sleep(1)
        cluster_data = cluster.status()

        expected_cluster_status_data_keys = [
            "env_servlet_processes",
            "server_pid",
            "runhouse_version",
            "cluster_config",
        ]

        actual_cluster_status_data_keys = list(cluster_data.keys())

        for key in expected_cluster_status_data_keys:
            assert key in actual_cluster_status_data_keys

        res = cluster_data.get("cluster_config")

        # test cluster config info
        assert res.get("server_port") == (cluster.server_port or DEFAULT_SERVER_PORT)
        assert res.get("server_connection_type") == cluster.server_connection_type
        assert res.get("den_auth") == cluster.den_auth
        assert res.get("resource_type") == cluster.RESOURCE_TYPE
        if res.get("resource_subtype") == "Cluster":
            assert res.get("ips") == cluster.ips
        else:
            assert res.get("compute_properties").get("ips") == cluster.ips

        assert process in cluster_data.get("env_servlet_processes").keys()
        assert "status_key1" in cluster_data.get("env_servlet_processes").get(
            process
        ).get("env_resource_mapping")
        assert {
            "resource_type": "str",
            "active_function_calls": [],
        } == cluster_data.get("env_servlet_processes").get(process).get(
            "env_resource_mapping"
        ).get(
            "status_key1"
        )
        sleep_calls = (
            cluster_data.get("env_servlet_processes")
            .get(process)
            .get("env_resource_mapping")
            .get("sleep_fn")
            .get("active_function_calls")
        )

        assert len(sleep_calls) == 3
        assert sleep_calls[0]["key"] == "sleep_fn"
        assert sleep_calls[0]["method_name"] == "call"
        assert sleep_calls[0]["request_id"]
        assert sleep_calls[0]["start_time"]

        # wait for threads to finish
        for call_thread in call_threads:
            call_thread.join()
        updated_status = cluster.status()
        # Check that the sleep calls are no longer active
        assert (
            updated_status.get("env_servlet_processes")
            .get(process)
            .get("env_resource_mapping")
            .get("sleep_fn")
            .get("active_function_calls")
            == []
        )

        # test memory usage info
        expected_servlet_keys = [
            "env_cpu_usage",
            "env_gpu_usage",
            "env_resource_mapping",
            "node_index",
            "node_ip",
            "node_name",
            "pid",
        ]
        process_names = list(cluster_data.get("env_servlet_processes").keys())
        process_names.sort()
        assert "env_servlet_processes" in cluster_data.keys()
        servlets_info = cluster_data.get("env_servlet_processes")
        actors_keys = list(servlets_info.keys())
        actors_keys.sort()
        assert process_names == actors_keys
        for process_name in process_names:
            servlet_info = servlets_info.get(process_name)
            servlet_info_keys = list(servlet_info.keys())
            servlet_info_keys.sort()
            assert servlet_info_keys == expected_servlet_keys

    @pytest.mark.skip("pending cluster.kill functionality")
    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_rh_status_pythonic_delete_process(self, cluster):
        process = cluster.ensure_process_created("process_name")
        summer_temp = rh.function(summer).to(system=cluster, process=process)
        call_summer_temp = summer_temp(1, 3)
        assert call_summer_temp == 4

        # make sure status is calculated properly before process deletion.
        self.test_rh_status_pythonic(cluster=cluster)

        cluster.delete(process)
        # make sure status is calculated properly after process deletion.
        self.test_rh_status_pythonic(cluster=cluster)

    def status_cli_test_logic(self, cluster, status_cli_command: str):
        default_process_name = DEFAULT_PROCESS_NAME

        cluster.put(key="status_key2", obj="status_value2")

        if cluster.ips == cluster.internal_ips != ["localhost"]:
            status_output_response = cluster.run_bash_over_ssh([status_cli_command])[0]
        else:
            status_output_response = cluster.run_bash([status_cli_command])[0]
        assert status_output_response[0] == 0
        status_output_string = status_output_response[1]
        # The string that's returned is utf-8 with the literal escape characters mixed in.
        # We need to convert the escape characters to their actual values to compare the strings.
        status_output_string = status_output_string.encode("utf-8").decode(
            "unicode_escape"
        )
        status_output_string = status_output_string.replace("\n", "")
        assert "Runhouse server is running" in status_output_string
        assert f"Runhouse v{rh.__version__}" in status_output_string
        assert f"server port: {cluster.server_port}" in status_output_string
        assert (
            f"server connection type: {cluster.server_connection_type}"
            in status_output_string
        )
        assert f"den auth: {str(cluster.den_auth)}" in status_output_string
        assert (
            f"resource subtype: {cluster.config().get('resource_subtype')}"
            in status_output_string
        )
        assert f"ips: {str(cluster.ips)}" in status_output_string
        assert "Serving " in status_output_string
        assert f"{default_process_name}" in status_output_string
        assert "status_key2 (str)" in status_output_string
        assert "creds" not in status_output_string

        # checking the memory info is printed correctly
        assert "CPU: " in status_output_string
        assert status_output_string.count("CPU: ") >= 1
        assert "pid: " in status_output_string
        assert status_output_string.count("pid: ") >= 1
        assert "node: " in status_output_string
        assert status_output_string.count("node: ") >= 1

        cloud_properties = cluster.config().get("compute_properties", None)
        if cloud_properties:
            properties_to_check = ["cloud", "instance_type", "region", "cost_per_hour"]
            for p in properties_to_check:
                property_value = cloud_properties.get(p)
                assert property_value in status_output_string

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_rh_status_cmd_with_no_den_ping(self, cluster):
        self.status_cli_test_logic(
            cluster=cluster, status_cli_command="runhouse cluster status"
        )

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_rh_status_cmd_with_den_ping(self, cluster):
        self.status_cli_test_logic(
            cluster=cluster, status_cli_command="runhouse cluster status --send-to-den"
        )

    @pytest.mark.skip("Restarting the server mid-test causes some errors, need to fix")
    @pytest.mark.level("local")
    @pytest.mark.clustertest
    # TODO: once fixed, extend this tests for gpu clusters as well.
    def test_rh_status_cli_not_in_cluster(self, cluster):
        default_process_name = DEFAULT_PROCESS_NAME

        cluster.put(key="status_key3", obj="status_value3")
        res = str(
            subprocess.check_output(
                ["runhouse", "cluster", "status", f"{cluster.name}"]
            ),
            "utf-8",
        )
        assert "😈 Runhouse server is running 🏃" in res
        assert f"server port: {cluster.server_port}" in res
        assert f"server connection_type: {cluster.server_connection_type}" in res
        assert f"den auth: {str(cluster.den_auth)}" in res
        assert f"resource subtype: {cluster.RESOURCE_TYPE.capitalize()}" in res
        assert f"ips: {str(cluster.ips)}" in res
        assert "Serving 🍦 :" in res
        assert f"{default_process_name} (runhouse.Env)" in res
        assert "status_key3 (str)" in res
        assert "ssh certs" not in res

    @pytest.mark.skip("Restarting the server mid-test causes some errors, need to fix")
    @pytest.mark.level("local")
    @pytest.mark.clustertest
    # TODO: once fixed, extend this tests for gpu clusters as well.
    def test_rh_status_stopped(self, cluster):
        try:
            cluster_name = cluster.name
            cluster.run_bash(["runhouse server stop"])
            res = subprocess.check_output(["runhouse", "status", cluster_name]).decode(
                "utf-8"
            )
            assert "Runhouse Daemon is not running" in res
            res = subprocess.check_output(
                ["runhouse", "cluster", "status", f"{cluster_name}_dont_exist"]
            ).decode("utf-8")
            error_txt = (
                f"Cluster {cluster_name}_dont_exist is not found in Den. Please save it, in order to get "
                f"its status"
            )
            assert error_txt in res
        finally:
            cluster.run_bash(["runhouse server restart"])

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_send_status_to_db(self, cluster):

        status = cluster.status()
        env_servlet_processes = status.pop("env_servlet_processes")
        status_data = {
            "daemon_status": RunhouseDaemonStatus.RUNNING,
            "resource_type": status.get("cluster_config").get("resource_type"),
            "resource_info": status,
            "env_servlet_processes": env_servlet_processes,
        }
        cluster_uri = rh.globals.rns_client.format_rns_address(cluster.rns_address)
        headers = rh.globals.rns_client.request_headers()
        api_server_url = rh.globals.rns_client.api_server_url
        post_status_data_resp = requests.post(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status",
            data=json.dumps(status_data),
            headers=headers,
        )
        assert post_status_data_resp.status_code in [200, 422]
        get_status_data_resp = requests.get(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status?limit=1",
            headers=headers,
        )
        assert get_status_data_resp.status_code == 200
        get_status_data = get_status_data_resp.json()["data"][0]
        assert get_status_data["resource_type"] == status.get("cluster_config").get(
            "resource_type"
        )
        assert get_status_data["daemon_status"] == RunhouseDaemonStatus.RUNNING

        assert get_status_data["resource_info"] == status
        for k in env_servlet_processes:
            if env_servlet_processes[k]["env_gpu_usage"] == {}:
                env_servlet_processes[k]["env_gpu_usage"] = {
                    "used_memory": None,
                    "utilization_percent": None,
                    "total_memory": None,
                }
        env_servlet_processes = sort_env_servlet_processes(env_servlet_processes)
        get_status_data["env_servlet_processes"] = sort_env_servlet_processes(
            get_status_data["env_servlet_processes"]
        )
        assert get_status_data["env_servlet_processes"] == env_servlet_processes

        status_data["daemon_status"] = RunhouseDaemonStatus.TERMINATED
        post_status_data_resp = requests.post(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status",
            data=json.dumps(status_data),
            headers=headers,
        )
        assert post_status_data_resp.status_code == 200
        get_status_data_resp = requests.get(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status?limit=1",
            headers=headers,
        )
        assert (
            get_status_data_resp.json()["data"][0]["daemon_status"]
            == RunhouseDaemonStatus.TERMINATED
        )

        # setting the status to running again, so it won't mess with the following tests
        # (when running all release suite at once, for example)
        post_status_data_resp = requests.post(
            f"{api_server_url}/resource/{cluster_uri}/cluster/status",
            data=json.dumps(status_data),
            headers=headers,
        )
        assert post_status_data_resp.status_code in [200, 422]

    ####################################################################################################
    # Default process tests
    ####################################################################################################

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_default_process_in_status(self, cluster):
        res = cluster.status()
        assert DEFAULT_PROCESS_NAME in res.get("env_servlet_processes")

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_put_in_default_process(self, cluster):
        k1 = get_random_str()
        cluster.put(k1, "v1")

        assert k1 in cluster.keys(process=DEFAULT_PROCESS_NAME)
        cluster.delete(k1)

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_fn_to_no_process_specified(self, cluster):
        remote_summer = rh.function(summer).to(cluster)

        found = False
        for p in cluster.list_processes():
            if p.startswith(remote_summer.name):
                found = True
                assert remote_summer.name in cluster.keys(process=p)
                break

        assert found

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_run_in_default_process(self, cluster):
        reqs = []
        if cluster.image:
            for step in cluster.image.setup_steps:
                if step.step_type == ImageSetupStepType.PACKAGES:
                    reqs += step.kwargs.get("reqs")
        for req in reqs:
            if isinstance(req, str) and "_" in req:
                # e.g. pytest_asyncio
                req = req.replace("_", "-")
                assert cluster.run_bash(f"pip freeze | grep {req}")[0] == 0

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_default_conda_env_created(self, cluster):
        if not cluster.image or not cluster.image.conda_env_name:
            pytest.skip("Default process is not in a conda env")

        assert cluster.image.conda_env_name in cluster.run_bash("conda info --envs")[1]

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_default_process_env_var_run(self, cluster):
        env_vars = {}
        if cluster.image:
            for setup_step in cluster.image.setup_steps:
                if setup_step.step_type == ImageSetupStepType.SET_ENV_VARS:
                    image_env_vars = _process_env_vars(
                        setup_step.kwargs.get("env_vars")
                    )
                    env_vars.update(image_env_vars)
        if not env_vars:
            pytest.skip("No env vars in cluster image")

        assert env_vars
        for var in env_vars.keys():
            res = cluster.run_bash([f"echo ${var}"])
            assert res[0][0] == 0
            assert env_vars[var] in res[0][1]

        get_env_var_cpu = rh.function(_get_env_var_value).to(system=cluster)
        for var in env_vars.keys():
            assert get_env_var_cpu(var) == env_vars[var]

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_run_within_cluster(self, cluster):
        if cluster.config().get("resource_subtype") == "Cluster":
            pytest.skip("run_bush don't work when running on the cluster, need to fix")
        remote_run = rh.function(run_in_no_env).to(cluster)
        res = remote_run("echo hello")

        if cluster.ips == cluster.internal_ips != ["localhost"]:
            exp = cluster.run_bash_over_ssh(["echo hello"])
        else:
            exp = cluster.run_bash(["echo hello"])

        # multinode case; check that the call passes on all nodes
        if len(cluster.ips) > 1:
            assert len(res) == len(exp)
            for node_id in range(len(res)):
                assert res[node_id][0] == 0
                assert res[node_id][1].strip() == exp[node_id][1].strip()
        else:
            # single-node case
            assert res[0] == 0
            assert res[1].strip() == exp[0][1].strip()

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_put_and_get(self, cluster):
        cluster._folder_mkdir(path="~/.rh/new-folder")
        folder_contents: list = cluster._folder_ls(path="~/.rh")
        assert "new-folder" in [os.path.basename(f) for f in folder_contents]

        cluster._folder_put(
            path="~/.rh/new-folder",
            contents={"sample.txt": "Hello World!"},
            overwrite=True,
        )

        file_contents = cluster._folder_get(path="~/.rh/new-folder/sample.txt")
        assert file_contents == "Hello World!"

        # Should not be able to put to an existing file unless `overwrite=True`
        with pytest.raises(ValueError):
            cluster._folder_put(
                path="~/.rh/new-folder",
                contents={"sample.txt": "Hello World!"},
            )

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_put_and_get_serialization_methods(self, cluster):
        from runhouse.servers.http.http_utils import deserialize_data

        raw_data = [1, 2, 3]
        pickle_serialization = "pickle"

        cluster._folder_put(
            path="~/.rh/new-folder",
            contents={"sample.pickle": raw_data},
            overwrite=True,
            serialization=pickle_serialization,
        )

        file_contents = cluster._folder_get(path="~/.rh/new-folder/sample.pickle")
        assert deserialize_data(file_contents, pickle_serialization) == raw_data

        json_serialization = "json"
        cluster._folder_put(
            path="~/.rh/new-folder",
            contents={"sample.text": raw_data},
            overwrite=True,
            serialization=json_serialization,
        )

        file_contents = cluster._folder_get(path="~/.rh/new-folder/sample.text")
        assert deserialize_data(file_contents, json_serialization) == raw_data

        with pytest.raises(AttributeError):
            # No serialization specified, default mode of "wb" used which is not supported for a list
            cluster._folder_put(
                path="~/.rh/new-folder",
                contents={"sample.pickle": raw_data},
                overwrite=True,
            )

        # with no serialization specified, but with "w" mode
        cluster._folder_put(
            path="~/.rh/new-folder",
            contents={"sample.txt": raw_data},
            overwrite=True,
            mode="w",
        )

        file_contents = cluster._folder_get(path="~/.rh/new-folder/sample.text")
        assert deserialize_data(file_contents, json_serialization) == raw_data

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_mkdir_mv_and_rm(self, cluster):
        cluster._folder_mkdir(path="~/.rh/new-folder")

        cluster._folder_mv(path="~/.rh/new-folder", dest_path="~/new-folder")
        file_contents = cluster._folder_ls(path="~")

        assert "new-folder" in [os.path.basename(f) for f in file_contents]

        # Should not be able to mv to an existing directory if `overwrite=False`
        cluster._folder_mkdir(path="~/.rh/another-new-folder")
        with pytest.raises(Exception):
            cluster._folder_mv(
                path="~/.rh/another-new-folder",
                dest_path="~/new-folder",
                overwrite=False,
            )

        # Delete folder contents and directory itself
        cluster._folder_rm(path="~/new-folder", recursive=True)

        assert not cluster._folder_exists(path="~/new-folder")

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_observability_enabled_by_default_on_cluster(self, cluster):
        # Disable observability locally, which will be reflected on the cluster once the server is restarted
        if cluster.image:
            rh.configs.disable_observability()
            cluster.restart_server()

            res = cluster.run_bash(["echo $disable_observability"])
            assert "True" in res[0][1]

    ####################################################################################################
    # Cluster list test
    ####################################################################################################
    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_default_pythonic(self, cluster):
        default_clusters = Cluster.list().get("den_clusters", {})
        assert len(default_clusters) > 0
        assert [
            den_cluster.get("Status") == ClusterStatus.RUNNING
            for den_cluster in default_clusters
        ]
        assert any(
            den_cluster
            for den_cluster in default_clusters
            if den_cluster.get("Name") == cluster.rns_address
        )

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_all_pythonic(self, cluster):
        # create dummy terminated cluster - set the daemon status to terminated, which will also
        # update the cluster status. Making sure that its name is in the same format as all testing clusters
        # (contains timestamp and uuid)
        terminated_cluster = rh.cluster(
            name=f"{create_folder_path()}-terminated-cluster",
            server_connection_type="ssh",
        ).save()
        set_daemon_and_cluster_status(
            terminated_cluster,
            daemon_status=RunhouseDaemonStatus.TERMINATED,
            cluster_status=ClusterStatus.TERMINATED,
        )

        all_clusters = Cluster.list(show_all=True).get("den_clusters", {})
        present_statuses = set(
            [den_cluster.get("Status") for den_cluster in all_clusters]
        )
        assert len(present_statuses) > 1
        assert ClusterStatus.RUNNING in present_statuses
        assert ClusterStatus.TERMINATED in present_statuses

        test_cluster = [
            den_cluster
            for den_cluster in all_clusters
            if den_cluster.get("Name") == cluster.rns_address
        ][0]
        assert test_cluster.get("Status") == ClusterStatus.RUNNING

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_status_pythonic(self, cluster):
        for status in [ClusterStatus.RUNNING, ClusterStatus.TERMINATED]:
            # check that filtered requests contains only specific status
            filtered_clusters = Cluster.list(status=status).get("den_clusters", {})
            if filtered_clusters:
                filtered_statuses = set(
                    [cluster.get("Status") for cluster in filtered_clusters]
                )
                assert status in filtered_statuses

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_since_pythonic(self, cluster):
        cluster.save()  # tls exposed local cluster is not saved by default

        minutes_time_filter = 10
        clusters = Cluster.list(since=f"{minutes_time_filter}m")
        recent_clusters = clusters.get("den_clusters", {})

        clusters_last_active_timestamps = set(
            [den_cluster.get("Last Active (UTC)") for den_cluster in recent_clusters]
        )

        assert len(clusters_last_active_timestamps) >= 1
        current_utc_time = datetime.utcnow().replace(tzinfo=timezone.utc)

        for timestamp in clusters_last_active_timestamps:
            # Convert the timestamp string to a naive datetime object
            timestamp_obj = datetime.strptime(timestamp, "%m/%d/%Y, %H:%M:%S")
            timestamp_obj = timestamp_obj.replace(tzinfo=timezone.utc)
            assert (
                current_utc_time - timestamp_obj
            ).total_seconds() <= minutes_time_filter * MINUTE

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_cmd_output_no_filters(self, capsys, cluster):
        import re
        import subprocess

        env = set_output_env_vars()

        process = subprocess.Popen(
            "runhouse cluster list",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=env,
        )
        process.wait()
        stdout = process.communicate()[0]
        capsys.readouterr()
        cmd_stdout = stdout.decode("utf-8")

        assert cmd_stdout

        # The output is printed as a table.
        # testing that the table name is printed correctly
        regex = f".*Clusters for {rh.configs.username}.*\(Running: .*/.*, Total Displayed: .*/.*\).*"
        assert re.search(regex, cmd_stdout)

        # testing that the table column names is printed correctly
        col_names = [
            "┃ Name",
            "┃ Cluster Type",
            "┃ Status",
            "┃ Last Active (UTC)",
            "┃ Autostop (Mins)",
        ]
        for name in col_names:
            assert name in cmd_stdout
        assert (
            f"Showing clusters that were active in the last {int(LAST_ACTIVE_AT_TIMEFRAME / HOUR)} hours."
            in cmd_stdout
        )
        assert cluster.rns_address in cmd_stdout

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_list_cmd_output_with_filters(self, capsys, cluster):

        import re
        import subprocess

        cluster.save()  # tls exposed local cluster is not saved by default

        subprocess_env = set_output_env_vars()

        for status in [ClusterStatus.RUNNING, ClusterStatus.TERMINATED]:
            process = subprocess.Popen(
                f"runhouse cluster list --status {status}",
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=subprocess_env,
            )
            process.wait()
            stdout = process.communicate()[0]
            capsys.readouterr()
            cmd_stdout = stdout.decode("utf-8")
            assert cmd_stdout

            # The output is printed as a table.
            # testing that the table name is printed correctly

            regex = ".*Clusters for.*\(Running: .*/.*, Total Displayed: .*/.*\).*"
            assert re.search(regex, cmd_stdout)

            # testing that the table column names is printed correctly
            col_names = [
                "┃ Name",
                "┃ Cluster Type",
                "┃ Status",
                "┃ Last Active (UTC)",
                "┃ Autostop (Mins)",
            ]
            for name in col_names:
                assert name in cmd_stdout

            assert (
                "Note: the above clusters have registered activity in the last 24 hours."
                not in cmd_stdout
            )

            if status == ClusterStatus.RUNNING:
                assert cluster.rns_address in cmd_stdout

            # Check other statuses not found in output
            cmd_stdout = cmd_stdout.replace("Running:", "")
            statuses = [s.lower() for s in list(ClusterStatus.__members__.keys())]
            statuses.remove(status)

            for status in statuses:
                assert status.capitalize() not in cmd_stdout

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_create_and_list_process(self, cluster):
        assert DEFAULT_PROCESS_NAME in cluster.list_processes()

        # create a process manually with the create_process functionality
        cluster.ensure_process_created(name="new_test_process_created_with_utility")
        assert "new_test_process_created_with_utility" in cluster.list_processes()

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_run_bash_in_process(self, cluster):
        process = cluster.ensure_process_created(name="test_process")
        res = cluster.run_bash("echo hello", process=process)
        assert res[0] == 0
        assert res[1].strip() == "hello"

    @pytest.mark.level("local")
    @pytest.mark.clustertest
    def test_cluster_kill_process(self, cluster):
        process = cluster.ensure_process_created(name="new_test_process")
        assert process in cluster.list_processes()
        cluster.put(key="new_key", obj="val", process=process)
        assert cluster.get("new_key") == "val"
        cluster.kill_process(process)
        assert cluster.get("new_key") is None
        assert process not in cluster.list_processes()



from enum import Enum
from typing import Any, Dict, List, Optional, Union

# Internal class to represent the image construction process
class ImageSetupStepType(Enum):
    """Enum for valid Image setup step types"""

    PACKAGES = "packages"
    CMD_RUN = "cmd_run"
    SETUP_CONDA_ENV = "setup_conda_env"
    RSYNC = "rsync"
    SYNC_SECRETS = "sync_secrets"
    SET_ENV_VARS = "set_env_vars"


class ImageSetupStep:
    def __init__(
        self,
        step_type: ImageSetupStepType,
        **kwargs: Dict[str, Any],
    ):
        """
        A component of the Runhouse Image, consisting of the step type (e.g. packages, set_env_vars),
        along with arguments to provide to the function corresponding to the step type.

        Args:
            step_type (ImageSetupStepType): Type of setup step used to provide the Image.
            kwargs (Dict[str, Any]): Please refer to the corresponding functions in ``Image`` to determine
                the correct keyword arguments to provide.
        """
        self.step_type = step_type
        self.kwargs = kwargs


class Image:
    def __init__(self, name: str = None, image_id: str = None):
        """
        Runhouse Image object, specifying cluster setup properties and steps.

        Args:
            name (str): Name to assign the Runhouse image.
            image_id (str): Machine image to use, if any. (Default: ``None``)

        Example:
            >>> my_image = (
            >>>     rh.Image(name="base_image")
            >>>     .setup_conda_env(
            >>>         conda_env_name="base_env",
            >>>         conda_config={"dependencies": ["python=3.11"], "name": "base_env"},
            >>>     )
            >>>     .install_packages(["numpy", "pandas"])
            >>>     .set_env_vars({"OMP_NUM_THREADS": 1})
            >>> )
        """

        self.name = name
        self.image_id = image_id

        self.setup_steps = []
        self.conda_env_name = None
        self.docker_secret = None

    @staticmethod
    def _setup_step_config(step: ImageSetupStep):
        """Get ImageSetupStep config"""
        config = {
            "step_type": step.step_type.value,
        }
        if step.step_type == ImageSetupStepType.SYNC_SECRETS:
            secrets = step.kwargs.get("providers")
            new_kwargs = []
            for secret in secrets:
                new_kwargs.append(
                    secret if isinstance(secret, str) else secret.config(values=False)
                )
            config["kwargs"] = new_kwargs
        else:
            config["kwargs"] = step.kwargs
        return config

    @staticmethod
    def _setup_step_from_config(step: Dict):
        """Convert setup step config (dict) to ImageSetupStep object"""
        step_type = step["step_type"]
        if step_type == "sync_secrets":
            from runhouse.resources.secrets.secret import Secret

            secrets = step["kwargs"]
            secret_list = []
            for secret in secrets:
                secret_list.append(
                    Secret.from_config(secret) if isinstance(secret, Dict) else secret
                )
            kwargs = {"providers": secret_list}
        else:
            kwargs = step["kwargs"]
        return ImageSetupStep(
            step_type=ImageSetupStepType(step_type),
            **kwargs,
        )

    def _save_sub_resources(self):
        secret_steps = [
            step
            for step in self.setup_steps
            if step.step_type == ImageSetupStepType.SYNC_SECRETS
        ]
        for step in secret_steps:
            from runhouse.resources.secrets.secret import Secret

            secrets = step.kwargs.get("providers")
            for secret in secrets:
                if isinstance(secret, Secret):
                    secret.save()

    def from_docker(self, image_id: str, docker_secret: Union["Secret", str] = None):
        """Set up and use an existing Docker image.

        Args:
            image_id (str): Docker image in the following format ``"<registry>/<image>:<tag>"``
            docker_secret (Secret or str, optional): Runhouse secret corresponding to information
                necessary to pull the image from a private registry, such as Docker Hub or cloud provider.
                See ``DockerRegistrySecret`` for more information.
        """
        if self.image_id:
            raise ValueError(
                "Setting both a machine image and docker image is not yet supported."
            )
        self.image_id = f"docker:{image_id}"
        self.docker_secret = docker_secret
        return self

    def config(self) -> Dict[str, Any]:
        config = {}
        if self.name:
            config["name"] = self.name
        if self.image_id:
            config["image_id"] = self.image_id
        if self.docker_secret:
            config["docker_secret"] = (
                self.docker_secret
                if isinstance(self.docker_secret, str)
                else self.docker_secret.config()
            )
        if self.setup_steps:
            config["setup_steps"] = [
                Image._setup_step_config(step) for step in self.setup_steps
            ]

        return config

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        img = Image(name=config.get("name"), image_id=config.get("image_id"))
        if config.get("setup_steps"):
            img.setup_steps = [
                Image._setup_step_from_config(step) for step in config["setup_steps"]
            ]

        docker_secret = config.get("docker_secret")
        if docker_secret:
            if isinstance(docker_secret, Dict):
                from runhouse.resources.secrets.secret import Secret

                docker_secret = Secret.from_config(docker_secret)
            img.docker_secret = docker_secret

        return img

    ########################################################
    # Steps to build the image
    ########################################################

    def install_packages(
        self, reqs: List[Union["Package", str]], conda_env_name: Optional[str] = None
    ):
        """Install the given packages.

        Args:
            reqs (List[Package or str]): List of packages to install on cluster and env.
            conda_env_name (str, optional): Name of conda env to install the package in, if relevant. If left empty,
                defaults to base environment. (Default: ``None``)
        """

        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.PACKAGES,
                reqs=reqs,
                conda_env_name=conda_env_name or self.conda_env_name,
            )
        )
        return self

    def run_bash(self, command: str, conda_env_name: Optional[str] = None):
        """Run bash commands.

        Args:
            command (str): Commands to run on the cluster.
            conda_env_name (str, optional): Name of conda env to run the command in, if applicable. (Defaut: ``None``)
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.CMD_RUN,
                command=command,
                conda_env_name=conda_env_name or self.conda_env_name,
            )
        )
        return self

    def setup_conda_env(self, conda_env_name: str, conda_config: Union[str, Dict]):
        """Setup Conda env

        Args:
            conda_env_name (str): Name of conda env to create.
            conda_config (str or Dict): Path or Dict referring to the conda yaml config to use to create the conda env.
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.SETUP_CONDA_ENV,
                conda_env_name=conda_env_name,
                conda_config=conda_config,
            )
        )
        self.conda_env_name = conda_env_name
        return self

    def rsync(
        self, source: str, dest: str, contents: bool = False, filter_options: str = None
    ):
        """Sync the contents of the source directory into the destination.

        Args:
            source (str): The source path.
            dest (str): The target path.
            contents (bool, optional): Whether the contents of the source directory or the directory
                itself should be copied to destination.
                If ``True`` the contents of the source directory are copied to the destination, and the source
                directory itself is not created at the destination.
                If ``False`` the source directory along with its contents are copied ot the destination, creating
                an additional directory layer at the destination. (Default: ``False``).
            filter_options (str, optional): The filter options for rsync.
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.RSYNC,
                source=source,
                dest=dest,
                contents=contents,
                filter_options=filter_options,
            )
        )

    def sync_secrets(self, providers: List[Union[str, "Secret"]]):
        """Send secrets for the given providers.

        Args:
            providers (List[str or Secret]): List of providers to send secrets for.
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.SYNC_SECRETS,
                providers=providers,
            )
        )
        return self

    def set_env_vars(self, env_vars: Union[str, Dict]):
        """Set environment variables.

        Args:
            env_vars (str or Dict): Dict of environment variables and values to set, or string pointing
                to local ``.env`` file consisting of env vars to set.
        """
        self.setup_steps.append(
            ImageSetupStep(
                step_type=ImageSetupStepType.SET_ENV_VARS,
                env_vars=env_vars,
            )
        )
        return self



from pathlib import Path
from typing import Optional, Union

from runhouse.logger import get_logger

from runhouse.resources.folders.folder import Folder
from runhouse.resources.hardware.utils import _get_cluster_from

logger = get_logger(__name__)


def folder(
    name: Optional[str] = None,
    path: Optional[Union[str, Path]] = None,
    system: Optional[Union[str, "Cluster"]] = None,
    load_from_den: bool = True,
    dryrun: bool = False,
) -> Folder:
    """Creates a Runhouse folder object, which can be used to interact with the folder at the given path.

    Args:
        name (Optional[str]): Name to give the folder, to be re-used later on.
        path (Optional[str or Path]): Path (or path) that the folder is located at.
        system (Optional[str or Cluster]): File system or cluster name. If prpre-oviding a file system this must be one of:
            [``file``, ``s3``, ``gs``].
        load_from_den (bool): Whether to try loading the Folder resource from Den. (Default: ``True``)
        dryrun (bool): Whether to create the Folder if it doesn't exist, or load a Folder object as a dryrun.
            (Default: ``False``)

    Returns:
        Folder: The resulting folder.

    Example:
        >>> rh.folder(name='training_imgs', path='remote_directory/images', system='s3').save()

        >>> # Load folder from above
        >>> reloaded_folder = rh.folder(name="training_imgs")
    """
    # TODO [DG] Include loud warning that relative paths are relative to the git root / working directory!

    if name and not any([path, system]):
        # If only the name is provided
        try:
            return Folder.from_name(name, load_from_den=load_from_den, dryrun=dryrun)
        except ValueError:
            # This is a rare instance where passing no constructor params is actually valid (anonymous folder),
            # so if we don't find the name, the user may still actually want to create a new folder.
            pass

    system = system or Folder.DEFAULT_FS
    if system == "s3":
        from .s3_folder import S3Folder

        logger.debug(f"Creating a S3 folder with name: {name}")
        return S3Folder(
            system=system,
            path=path,
            name=name,
            dryrun=dryrun,
        )
    elif system == "gs":
        from .gcs_folder import GCSFolder

        logger.debug(f"Creating a GS folder with name: {name}")
        return GCSFolder(
            system=system,
            path=path,
            name=name,
            dryrun=dryrun,
        )

    if system == Folder.DEFAULT_FS:
        logger.debug(f"Creating local folder with name: {name}")
        return Folder(
            system=system,
            path=path,
            name=name,
            dryrun=dryrun,
        )

    from runhouse import Cluster

    cluster_system = _get_cluster_from(system, dryrun=dryrun)
    if isinstance(cluster_system, Cluster):
        logger.debug(f"Creating folder {name} for cluster: {cluster_system.name}")
        return Folder(
            system=cluster_system,
            path=path,
            name=name,
            dryrun=dryrun,
        )

    raise ValueError(
        f"System '{system}' not currently supported. If the file system "
        f"is a cluster (ex: '/my-user/rh-cpu'), make sure the cluster config has been saved."
    )



import time

import requests

from runhouse.globals import rns_client
from runhouse.logger import get_logger

logger = get_logger(__name__)


def profile(func, reps=10):
    times = []
    for _ in range(reps):
        start = time.time()
        assert func()
        times.append(round((time.time() - start) * 1000, 2))
    return times, sum(times) / len(times)


def run_performance_tests(summer_func):
    cluster = summer_func.system
    times_list, avg_time = profile(lambda: summer_func.system.keys() is not None)
    print(f"Listing keys took {round(avg_time, 2)} ms: {times_list}")

    times_list, avg_time = profile(lambda: summer_func(1, 5) == 6)
    print(f"Call with logs took {round(avg_time, 2)} ms: {times_list}")

    times_list, avg_time = profile(lambda: summer_func(1, 5, stream_logs=False) == 6)
    print(f"Call without logs took {round(avg_time, 2)} ms: {times_list}")

    port = cluster.client.port
    suffix = "https" if cluster._use_https else "http"
    address = cluster.server_address

    call_url = f"{suffix}://{address}:{port}/summer_func/call/?serialization=None"
    logger.info(f"Call url: {call_url}")
    times_list, avg_time = profile(
        lambda: requests.post(
            call_url,
            json={"args": [1, 2]},
            headers=rns_client.request_headers(cluster.rns_address)
            if cluster.den_auth
            else None,
            verify=cluster.client.verify,
        ).json()
        == 3
    )
    print(f"{suffix} call took {round(avg_time, 2)} ms: {times_list}")


def test_roundtrip_performance(summer_func):
    run_performance_tests(summer_func)


def test_https_roundtrip_performance(summer_func_with_auth):
    run_performance_tests(summer_func_with_auth)



import configparser
import copy
import os

from typing import Dict

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class AzureSecret(ProviderSecret):
    """
    .. note::
            To create an AzureSecret, please use the factory method :func:`provider_secret` with ``provider="azure"``.
    """

    # values format: {"subscription_id": subscription_id}
    _PROVIDER = "azure"
    _DEFAULT_CREDENTIALS_PATH = "~/.azure/clouds.config"
    _DEFAULT_ENV_VARS = {"subscription_id": "AZURE_SUBSCRIPTION_ID"}

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return AzureSecret(**config, dryrun=dryrun)

    def _write_to_file(
        self,
        path: str = None,
        values: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            subscription_id = values["subscription_id"]

            parser = configparser.ConfigParser()
            section_name = "AzureCloud"
            parser.add_section(section_name)
            parser.set(
                section=section_name,
                option="subscription",
                value=subscription_id,
            )

            full_path = create_local_dir(path)
            with open(full_path, "w") as f:
                parser.write(f)

            if write_config:
                new_secret._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _from_path(self, path: str = None):
        config = configparser.ConfigParser()
        if path and os.path.exists(os.path.expanduser(path)):
            path = os.path.expanduser(path)
            config.read(path)
        if config and "AzureCloud" in config.sections():
            subscription_id = config["AzureCloud"]["subscription"]
            return {"subscription_id": subscription_id}
        return {}



import copy
import json
import os
from pathlib import Path

from typing import Dict

from runhouse.resources.secrets.provider_secrets.provider_secret import ProviderSecret
from runhouse.resources.secrets.utils import _check_file_for_mismatches
from runhouse.utils import create_local_dir


class GCPSecret(ProviderSecret):
    """
    .. note::
            To create a GCPSecret, please use the factory method :func:`provider_secret` with ``provider="gcp"``.
    """

    _PROVIDER = "gcp"
    _DEFAULT_CREDENTIALS_PATH = "~/.config/gcloud/application_default_credentials.json"
    _DEFAULT_ENV_VARS = {
        "client_id": "CLIENT_ID",
        "client_secret": "CLIENT_SECRET",
    }

    @staticmethod
    def from_config(config: dict, dryrun: bool = False, _resolve_children: bool = True):
        return GCPSecret(**config, dryrun=dryrun)

    def _write_to_file(
        self,
        path: str,
        values: Dict = None,
        overwrite: bool = False,
        write_config: bool = True,
    ):
        new_secret = copy.deepcopy(self)
        if not _check_file_for_mismatches(
            path, self._from_path(path), values, overwrite
        ):
            Path(path).parent.mkdir(parents=True, exist_ok=True)

            full_path = create_local_dir(path)
            with open(full_path, "w+") as f:
                json.dump(values, f, indent=4)

            if write_config:
                new_secret._add_to_rh_config(path)

        new_secret._values = None
        new_secret.path = path
        return new_secret

    def _from_path(self, path: str = None):
        config = {}
        if path and os.path.exists(os.path.expanduser(path)):
            with open(os.path.expanduser(path), "r") as config_file:
                config = json.load(config_file)
        return config



import json
import os
import uuid

import pytest

from runhouse.constants import DEFAULT_PROCESS_NAME
from runhouse.globals import rns_client
from runhouse.resources.resource import Resource
from runhouse.servers.http.http_utils import (
    DeleteObjectParams,
    deserialize_data,
    PutObjectParams,
    PutResourceParams,
    RenameObjectParams,
    serialize_data,
)

from tests.utils import friend_account

INVALID_HEADERS = {"Authorization": "Bearer InvalidToken"}

# Helper used for testing rh.Function
def summer(a, b):
    return a + b


@pytest.mark.servertest
@pytest.mark.usefixtures("cluster")
class TestHTTPServerDocker:
    """
    Test the HTTP server running on a local Docker container.

    These will be tested on a container with Den Auth enabled, and a container without.
    """

    UNIT = {
        "cluster": [
            "docker_cluster_pk_ssh_den_auth",
            "docker_cluster_pk_ssh_no_auth",
        ]
    }
    LOCAL = {
        "cluster": [
            "docker_cluster_pk_ssh_den_auth",
            "docker_cluster_pk_ssh_no_auth",
            "docker_cluster_pk_http_exposed",  # Represents within VPC use case
        ]
    }

    @pytest.mark.level("local")
    def test_check_server(self, http_client):
        response = http_client.get("/check")
        assert response.status_code == 200

    @pytest.mark.level("local")
    def test_put_resource(self, http_client, cluster):
        state = None
        resource = Resource(name="test_resource3", system=cluster)
        data = serialize_data(
            (resource.config(condensed=False), state, resource.dryrun), "pickle"
        )
        response = http_client.post(
            "/resource",
            json=PutResourceParams(
                serialized_data=data, serialization="pickle"
            ).model_dump(),
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

    @pytest.mark.level("local")
    def test_put_object_and_get_keys(self, http_client, cluster):
        key = "key1"
        test_list = list(range(5, 50, 2)) + ["a string"]
        response = http_client.post(
            "/object",
            json=PutObjectParams(
                key=key,
                serialized_data=serialize_data(test_list, "pickle"),
                serialization="pickle",
            ).model_dump(),
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.get(
            "/keys", headers=rns_client.request_headers(cluster.rns_address)
        )
        assert response.status_code == 200
        assert key in response.json().get("data")

    @pytest.mark.level("local")
    def test_rename_object(self, http_client, cluster):
        old_key = "key1"
        new_key = "key2"
        response = http_client.post(
            "/rename",
            json=RenameObjectParams(key=old_key, new_key=new_key).model_dump(),
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.get(
            "/keys", headers=rns_client.request_headers(cluster.rns_address)
        )
        assert new_key in response.json().get("data")

    @pytest.mark.level("local")
    def test_delete_obj(self, http_client, cluster):
        # https://www.python-httpx.org/compatibility/#request-body-on-http-methods
        key = "key2"
        response = http_client.post(
            url="/delete_object",
            json=DeleteObjectParams(keys=[key]).model_dump(),
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.get(
            "/keys", headers=rns_client.request_headers(cluster.rns_address)
        )
        assert key not in response.json().get("data")

    # TODO test get_call, refactor into proper fixtures
    @pytest.mark.level("local")
    def test_call_module_method(self, http_client, remote_func):
        # Create new func on the cluster, then call it
        method_name = "call"
        module_name = remote_func.name
        data = {"args": [1, 2], "kwargs": {}}

        response = http_client.post(
            f"{module_name}/{method_name}",
            json={
                "data": serialize_data(data, "pickle"),
                "stream_logs": True,
                "serialization": "pickle",
                "run_name": "test_call_module_method",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 200

        resp_obj: dict = json.loads(response.text.split("\n")[0])

        if resp_obj["output_type"] == "stdout":
            assert (
                f"{DEFAULT_PROCESS_NAME} env: Calling method call on module summer\n"
                in resp_obj["data"][0]
            )

        if resp_obj["output_type"] == "result_serialized":
            assert deserialize_data(resp_obj["data"], resp_obj["serialization"]) == 3

    @pytest.mark.level("local")
    def test_call_module_method_get_call(self, http_client, remote_func):
        method_name = "call"
        module_name = remote_func.name

        response = http_client.get(
            f"{module_name}/{method_name}",
            params={
                "a": 1,
                "b": 2,
                "run_name": "test_call_module_method_get_call",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 200

        resp_obj: dict = response.json()

        assert resp_obj["output_type"] == "result_serialized"

        # Right now everything via a GET request is serialized as strings
        assert deserialize_data(resp_obj["data"], resp_obj["serialization"]) == "12"

    @pytest.mark.level("local")
    def test_log_streaming_call(self, http_client, remote_log_streaming_func):
        # Create new func on the cluster, then call it
        method_name = "call"
        module_name = remote_log_streaming_func.name
        clus = remote_log_streaming_func.system
        data = {"args": [3], "kwargs": {}}

        url = f"{clus.endpoint()}/{module_name}/{method_name}"

        with http_client.stream(
            "POST",
            url,
            json={
                "data": serialize_data(data, "pickle"),
                "stream_logs": True,
                "serialization": "pickle",
                "run_name": "test_log_streaming_call",
            },
            headers=rns_client.request_headers(clus.rns_address),
        ) as r:
            assert r.status_code == 200
            for res in r.iter_lines():
                resp_obj: dict = json.loads(res)

                # Might be too aggressive to assert the exact print order and timing, but for now this works
                if resp_obj["output_type"] == "stdout":
                    if "env" in resp_obj["data"][0]:
                        assert (
                            f"{DEFAULT_PROCESS_NAME} env: Calling method call on module do_printing_and_logging\n"
                            in resp_obj["data"][0]
                        )
                    else:
                        assert "Hello from the cluster stdout!" in resp_obj["data"][0]
                        assert "Hello from the cluster logs!" in resp_obj["data"][1]

                if resp_obj["output_type"] == "result_serialized":
                    assert (
                        deserialize_data(resp_obj["data"], resp_obj["serialization"])
                        == 3
                    )

    @pytest.mark.level("local")
    def test_folder_put_and_get_and_mv(self, http_client, cluster):
        file_name = str(uuid.uuid4())
        response = http_client.post(
            "/folder/method/put",
            json={
                "path": "~/.rh",
                "contents": {f"{file_name}.txt": "Hello, world!"},
                "overwrite": True,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/get",
            json={"path": f"~/.rh/{file_name}.txt"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        resp_json = response.json()

        assert resp_json["serialization"] is None
        assert resp_json["output_type"] == "result_serialized"
        assert resp_json["data"] == "Hello, world!"

        dest_path = f"~/{file_name}.txt"
        response = http_client.post(
            "/folder/method/mv",
            json={
                "path": f"~/.rh/{file_name}.txt",
                "dest_path": dest_path,
                "overwrite": True,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/get",
            json={"path": dest_path},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200
        assert response.json()["data"] == "Hello, world!"

    @pytest.mark.level("local")
    def test_folder_content_serialization_methods(self, http_client, cluster):
        file_name = str(uuid.uuid4())

        # Save data with "pickle" serialization
        pickle_serialization = "pickle"
        response = http_client.post(
            "/folder/method/put",
            json={
                "path": "~/.rh",
                "contents": serialize_data(
                    data={f"{file_name}.txt": "Hello, world!"},
                    serialization=pickle_serialization,
                ),
                "serialization": pickle_serialization,
                "overwrite": True,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        # Result should be returned as pickled data
        response = http_client.post(
            "/folder/method/get",
            json={"path": f"~/.rh/{file_name}.txt"},
            headers=rns_client.request_headers(cluster.rns_address),
        )

        resp_json = response.json()
        assert resp_json["serialization"] is None
        assert resp_json["output_type"] == "result_serialized"
        assert (
            deserialize_data(resp_json["data"], pickle_serialization) == "Hello, world!"
        )

        # Save data with "json" serialization
        json_serialization = "json"
        response = http_client.post(
            "/folder/method/put",
            json={
                "path": "~/.rh",
                "contents": serialize_data(
                    {"new_file.txt": "Hello, world!"}, json_serialization
                ),
                "serialization": json_serialization,
                "overwrite": True,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/get",
            json={"path": "~/.rh/new_file.txt"},
            headers=rns_client.request_headers(cluster.rns_address),
        )

        resp_json = response.json()
        assert resp_json["serialization"] is None
        assert resp_json["output_type"] == "result_serialized"
        assert (
            deserialize_data(resp_json["data"], json_serialization) == "Hello, world!"
        )

        # Save data with no serialization
        response = http_client.post(
            "/folder/method/put",
            json={
                "path": "~/.rh",
                "contents": {"new_file.txt": "Hello, world!"},
                "overwrite": True,
                "serialization": None,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/get",
            json={"path": "~/.rh/new_file.txt"},
            headers=rns_client.request_headers(cluster.rns_address),
        )

        resp_json = response.json()
        assert resp_json["serialization"] is None
        assert resp_json["output_type"] == "result_serialized"
        assert resp_json["data"] == "Hello, world!"

    @pytest.mark.level("local")
    def test_folder_put_pickle_object(self, http_client, cluster):
        file_name = str(uuid.uuid4())

        raw_data = [1, 2, 3]
        serialization = "pickle"
        serialized_contents = serialize_data(
            {f"{file_name}.pickle": raw_data}, serialization=serialization
        )

        # need to specify the serialization method here
        response = http_client.post(
            "/folder/method/put",
            json={
                "path": "~/.rh",
                "contents": serialized_contents,
                "overwrite": True,
                "serialization": serialization,
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/get",
            json={"path": f"~/.rh/{file_name}.pickle"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        resp_json = response.json()
        assert resp_json["output_type"] == "result_serialized"
        assert deserialize_data(resp_json["data"], serialization) == raw_data

    @pytest.mark.level("local")
    def test_folder_mkdir_rm_and_ls(self, http_client, cluster):
        response = http_client.post(
            "/folder/method/mkdir",
            json={"path": "~/.rh/new-folder"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        # delete the file
        response = http_client.post(
            "/folder/method/rm",
            json={
                "path": "~/.rh/new-folder",
                "contents": ["new_file.txt"],
            },
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        # empty folder should still be there since recursive not explicitly set to `True`
        response = http_client.post(
            "/folder/method/ls",
            json={"path": "~/.rh"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        file_names: list = response.json().get("data")
        assert "new-folder" in [os.path.basename(f) for f in file_names]

        # Delete the now empty folder
        response = http_client.post(
            "/folder/method/rm",
            json={"path": "~/.rh/new-folder"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        response = http_client.post(
            "/folder/method/ls",
            json={"path": "~/.rh"},
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        file_names: list = response.json().get("data")
        assert "new-folder" not in [os.path.basename(f) for f in file_names]

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_async_call(self, async_http_client, remote_func):
        method = "call"

        response = await async_http_client.post(
            f"/{remote_func.name}/{method}",
            json={
                "data": {
                    "args": [1, 2],
                    "kwargs": {},
                },
                "serialization": None,
                "run_name": "test_async_call",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 200
        assert response.json() == 3

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_async_call_with_invalid_serialization(
        self, async_http_client, remote_func
    ):
        method = "call"

        response = await async_http_client.post(
            f"/{remote_func.name}/{method}",
            json={
                "data": {
                    "args": [1, 2],
                    "kwargs": {},
                },
                "serialization": "random",
                "run_name": "test_async_call_with_invalid_serialization",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 400
        assert "Invalid serialization type" in response.text

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_async_call_with_pickle_serialization(
        self, async_http_client, remote_func
    ):
        method = "call"

        response = await async_http_client.post(
            f"/{remote_func.name}/{method}",
            json={
                "data": serialize_data(
                    {
                        "args": [1, 2],
                        "kwargs": {},
                    },
                    "pickle",
                ),
                "serialization": "pickle",
                "run_name": "test_async_call_with_pickle_serialization",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 200
        assert (
            deserialize_data(response.json()["data"], response.json()["serialization"])
            == 3
        )

    @pytest.mark.level("local")
    @pytest.mark.asyncio
    async def test_async_call_with_json_serialization(
        self, async_http_client, remote_func
    ):
        method = "call"

        response = await async_http_client.post(
            f"/{remote_func.name}/{method}",
            json={
                "data": json.dumps(
                    {
                        "args": [1, 2],
                        "kwargs": {},
                    }
                ),
                "serialization": "json",
                "run_name": "test_async_call_with_json_serialization",
            },
            headers=rns_client.request_headers(remote_func.system.rns_address),
        )
        assert response.status_code == 200
        assert response.json()["data"] == "3"

    @pytest.mark.level("local")
    def test_valid_cluster_token(self, http_client, cluster):
        """User who created the cluster should be able to call cluster APIs with their default Den token
        or with a cluster subtoken."""
        response = http_client.get(
            "/keys",
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

        # Use cluster subtoken with resource address as the current username (cluster owner)
        response = http_client.get(
            "/keys",
            headers=rns_client.request_headers(rns_client.username),
        )
        assert response.status_code == 200

        # Use cluster subtoken with resource address as the cluster's rns address (/cluster_owner/cluster_name)
        response = http_client.get(
            "/keys",
            headers=rns_client.request_headers(cluster.rns_address),
        )
        assert response.status_code == 200

    @pytest.mark.level("local")
    def test_invalid_cluster_token(self, http_client, cluster):
        """Invalid cluster token should not be able to access the cluster APIs if den auth is enabled."""
        subtoken = f"abcdefg123+{rns_client.username}+{rns_client.username}"
        response = http_client.get(
            "/keys",
            headers={"Authorization": f"Bearer {subtoken}"},
        )

        if cluster.den_auth:
            # Should not be able to access cluster APIs
            assert response.status_code == 403
            assert "Cluster access is required for this operation." in response.text
        else:
            # If den auth is not enabled token should be ignored
            assert response.status_code == 200

    @pytest.mark.level("local")
    def test_no_access_with_invalid_default_token(self, http_client, cluster):
        """Invalid bearer token should result in a 401 if den auth is enabled."""
        invalid_subtoken = "invalid_resource_address"
        response = http_client.get(
            "/keys",
            headers={"Authorization": f"Bearer {invalid_subtoken}"},
        )

        if cluster.den_auth:
            assert response.status_code == 403
        else:
            assert response.status_code == 200


@pytest.mark.servertest
@pytest.mark.usefixtures("cluster")
class TestHTTPServerDockerDenAuthOnly:
    """
    Testing HTTP Server, but only against a container with Den Auth enabled.

    TODO: What is the expected behavior if there are invalid headers,
    but it is a server without Den Auth enabled at all?
    """

    UNIT = {"cluster": ["docker_cluster_pk_ssh_den_auth"]}
    LOCAL = {"cluster": ["docker_cluster_pk_ssh_den_auth"]}
    MINIMAL = {"cluster": ["docker_cluster_pk_ssh_den_auth"]}
    RELEASE = {"cluster": ["docker_cluster_pk_ssh_den_auth"]}
    MAXIMAL = {"cluster": ["docker_cluster_pk_ssh_den_auth"]}

    # -------- INVALID TOKEN / CLUSTER ACCESS TESTS ----------- #
    @pytest.mark.level("local")
    def test_no_access_to_cluster(self, http_client, cluster):
        # Make sure test account doesn't have access to the cluster (created by logged-in account, Den Tester in CI)
        cluster.revoke(["info@run.house"])
        cluster.enable_den_auth(flush=True)  # Flush auth cache

        import requests

        with friend_account():
            # Test accounts with Den auth are created under test_account
            res = requests.get(
                f"{rns_client.api_server_url}/resource",
                headers=rns_client.request_headers(cluster.rns_address),
            )
            assert cluster.rns_address not in [
                config["name"] for config in res.json().get("data", [])
            ]
            response = http_client.get(
                "/keys", headers=rns_client.request_headers(cluster.rns_address)
            )

        assert response.status_code == 403
        assert "Cluster access is required for this operation." in response.text

    @pytest.mark.level("local")
    def test_request_with_no_token(self, http_client):
        response = http_client.get("/keys")  # No headers are passed
        assert response.status_code == 401

        assert "No Runhouse token provided." in response.text

    @pytest.mark.level("local")
    def test_check_server_with_invalid_token(self, http_client):
        response = http_client.get("/check", headers=INVALID_HEADERS)
        # Should be able to ping the server even without a valid token
        assert response.status_code == 200

    @pytest.mark.level("local")
    def test_put_resource_with_invalid_token(self, http_client, cluster):
        state = None
        resource = Resource(name="test_resource1", system=cluster)
        data = serialize_data(
            (resource.config(condensed=False), state, resource.dryrun), "pickle"
        )
        response = http_client.post(
            "/resource",
            json=PutResourceParams(
                serialized_data=data, serialization="pickle"
            ).model_dump(),
            headers=INVALID_HEADERS,
        )
        assert response.status_code == 403
        assert "Cluster access is required for this operation." in response.text

    @pytest.mark.level("local")
    def test_call_module_method_with_invalid_token(self, http_client, remote_func):
        method_name = "call"
        module_name = remote_func.name
        args = (1, 2)
        kwargs = {}

        response = http_client.post(
            f"{module_name}/{method_name}",
            json={
                "data": {"args": args, "kwargs": kwargs},
                "stream_logs": False,
                "serialization": None,
                "run_name": "test_call_module_method_with_invalid_token",
            },
            headers=INVALID_HEADERS,
        )
        assert response.status_code == 403
        assert "Unauthorized access to resource summer" in response.text

    @pytest.mark.level("local")
    def test_put_object_with_invalid_token(self, http_client):
        test_list = list(range(5, 50, 2)) + ["a string"]
        response = http_client.post(
            "/object",
            json=PutObjectParams(
                key="key1",
                serialized_data=serialize_data(test_list, "pickle"),
                serialization="pickle",
            ).model_dump(),
            headers=INVALID_HEADERS,
        )
        assert response.status_code == 403
        assert "Cluster access is required for this operation." in response.text

    @pytest.mark.level("local")
    def test_rename_object_with_invalid_token(self, http_client):
        old_key = "key1"
        new_key = "key2"
        response = http_client.post(
            "/rename",
            json=RenameObjectParams(key=old_key, new_key=new_key).model_dump(),
            headers=INVALID_HEADERS,
        )
        assert response.status_code == 403
        assert "Cluster access is required for this operation." in response.text

    @pytest.mark.level("local")
    def test_get_keys_with_invalid_token(self, http_client):
        response = http_client.get("/keys", headers=INVALID_HEADERS)

        assert response.status_code == 403
        assert "Cluster access is required for this operation." in response.text


@pytest.fixture(scope="function")
def client(request):
    return request.getfixturevalue(request.param)


@pytest.mark.servertest
@pytest.mark.usefixtures("setup_cluster_config")
class TestHTTPServerNoDocker:
    """
    Directly analogous to the Docker equivalent above, but with a fully
    local server instead of Docker.

    TODO (RB+JL): This class should really be
    combined with the Docker test class above using some pytest magic.
    """

    # There is no default for this fixture, we should specify
    # it for each testing level.
    UNIT = {"client": ["local_client", "local_client_with_den_auth"]}
    LOCAL = {"client": ["local_client", "local_client_with_den_auth"]}
    MINIMAL = {"client": ["local_client", "local_client_with_den_auth"]}
    RELEASE = {"client": ["local_client", "local_client_with_den_auth"]}
    MAXIMAL = {"client": ["local_client", "local_client_with_den_auth"]}

    @pytest.mark.level("unit")
    def test_check_server(self, client):
        response = client.get("/check")
        assert response.status_code == 200

    @pytest.mark.level("unit")
    def test_put_resource(self, client, local_cluster):
        resource = Resource(name="local-resource")
        state = None
        data = serialize_data(
            (resource.config(condensed=False), state, resource.dryrun), "pickle"
        )
        response = client.post(
            "/resource",
            json=PutResourceParams(
                serialized_data=data, serialization="pickle"
            ).model_dump(),
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert response.status_code == 200

    @pytest.mark.level("unit")
    def test_put_object(self, client, local_cluster):
        test_list = list(range(5, 50, 2)) + ["a string"]
        response = client.post(
            "/object",
            json=PutObjectParams(
                key="key1",
                serialized_data=serialize_data(test_list, "pickle"),
                serialization="pickle",
            ).model_dump(),
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert response.status_code == 200

    @pytest.mark.level("unit")
    def test_rename_object(self, client, local_cluster):
        old_key = "key1"
        new_key = "key2"
        response = client.post(
            "/rename",
            json=RenameObjectParams(key=old_key, new_key=new_key).model_dump(),
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert response.status_code == 200

        response = client.get(
            "/keys",
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert new_key in response.json().get("data")

    @pytest.mark.level("unit")
    def test_get_keys(self, client, local_cluster):
        response = client.get(
            "/keys",
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert response.status_code == 200
        assert "key2" in response.json().get("data")

    @pytest.mark.level("unit")
    def test_delete_obj(self, client, local_cluster):
        key = "key"
        response = client.post(
            url="/delete_object",
            json=DeleteObjectParams(keys=[key]).model_dump(),
            headers=rns_client.request_headers(local_cluster.rns_address),
        )
        assert response.status_code == 200

        response = client.get(
            "/keys", headers=rns_client.request_headers(local_cluster.rns_address)
        )
        assert key not in response.json().get("data")

    # TODO [JL]: Test call_module_method and async_call with local and not just Docker.


@pytest.mark.servertest
@pytest.mark.usefixtures("setup_cluster_config")
class TestHTTPServerNoDockerDenAuthOnly:
    """
    Directly analogous to the Docker equivalent above, but with a fully
    local server instead of Docker.

    TODO (RB+JL): This class should really be
    combined with the Docker test class above using some pytest magic.
    """

    # -------- INVALID TOKEN / CLUSTER ACCESS TESTS ----------- #
    @pytest.mark.level("unit")
    def test_request_with_no_token(self, local_client_with_den_auth):
        response = local_client_with_den_auth.get(
            "/keys", headers={"Authorization": ""}
        )  # No headers are passed
        assert response.status_code == 401

        assert "No Runhouse token provided." in response.text

    @pytest.mark.level("unit")
    def test_check_server_with_invalid_token(self, local_client_with_den_auth):
        response = local_client_with_den_auth.get("/check", headers=INVALID_HEADERS)
        # Should be able to ping the server even without a valid token
        assert response.status_code == 200

    @pytest.mark.level("unit")
    def test_put_resource_with_invalid_token(self, local_client_with_den_auth):
        resource = Resource(name="test_resource2")
        state = None
        data = serialize_data(
            (resource.config(condensed=False), state, resource.dryrun), "pickle"
        )
        resp = local_client_with_den_auth.post(
            "/resource",
            json=PutResourceParams(
                serialized_data=data, serialization="pickle"
            ).model_dump(),
            headers=INVALID_HEADERS,
        )

        assert resp.status_code == 403

    @pytest.mark.level("unit")
    def test_put_object_with_invalid_token(self, local_client_with_den_auth):
        test_list = list(range(5, 50, 2)) + ["a string"]
        resp = local_client_with_den_auth.post(
            "/object",
            json=PutObjectParams(
                key="key1",
                serialized_data=serialize_data(test_list, "pickle"),
                serialization="pickle",
            ).model_dump(),
            headers=INVALID_HEADERS,
        )
        assert resp.status_code == 403

    @pytest.mark.level("unit")
    def test_rename_object_with_invalid_token(self, local_client_with_den_auth):
        old_key = "key1"
        new_key = "key2"
        resp = local_client_with_den_auth.post(
            "/rename",
            json=RenameObjectParams(key=old_key, new_key=new_key).model_dump(),
            headers=INVALID_HEADERS,
        )
        assert resp.status_code == 403

    @pytest.mark.level("unit")
    def test_get_keys_with_invalid_token(self, local_client_with_den_auth):
        resp = local_client_with_den_auth.get("/keys", headers=INVALID_HEADERS)
        assert resp.status_code == 403

    # TODO (JL): Test call_module_method.



# Package tests mostly migrated to tests/test_resources/test_data/test_package.py

import subprocess
from pathlib import Path

import runhouse as rh

extra_index_url = "--extra-index-url https://pypi.python.org/simple/"
cuda_116_url = "--index-url https://download.pytorch.org/whl/cu116"


def send_tensor_to_cuda():
    try:
        import torch

        a = torch.LongTensor(1).random_(0, 10)
        a = a.to(device="cuda")
        return isinstance(a, torch.Tensor)

    except Exception as e:
        return e


def test_mount_local_package_to_cluster(cluster):
    mount_path = "package_mount"
    package = rh.Package.from_string("./").to(cluster, path=mount_path)

    assert isinstance(package.install_target, rh.Folder)
    assert package.install_target.system == cluster
    assert mount_path in cluster.run_bash(["ls"])[0][1]


def test_package_file_system_to_cluster(cluster, s3_package):
    assert s3_package.install_target.system == "s3"
    assert s3_package.install_target.exists_in_system()

    folder_name = Path(s3_package.install_target.path).stem
    s3_package.to(system=cluster, path=folder_name)

    # Confirm the package's folder is now on the cluster
    assert "sample_file_0.txt" in cluster.run_bash([f"ls {folder_name}"])[0][1]


def test_getting_cuda_version_on_clusters(request, ondemand_cluster):
    """Gets the cuda version on the cluster and asserts it is the expected version"""
    return_codes: list = ondemand_cluster.run_python(
        [
            "from runhouse.resources.hardware.utils import detect_cuda_version_or_cpu",
            "print(detect_cuda_version_or_cpu())",
        ]
    )
    cuda_version_or_cpu = return_codes[0][1].strip().split("\n")[-1]
    print(f"{ondemand_cluster.name}: {cuda_version_or_cpu}")

    instance_type = ondemand_cluster.instance_type.lower()

    # Save the cuda version (or indicate cpu) for each cluster
    request.config.cache.set(instance_type, cuda_version_or_cpu)

    if instance_type.startswith("k80"):
        assert cuda_version_or_cpu == "11.3"
    elif instance_type.startswith("cpu"):
        assert cuda_version_or_cpu == "cpu"
    else:
        assert cuda_version_or_cpu == "11.7"


def test_install_cmd_for_torch_on_cluster(request, ondemand_cluster):
    """Checks that the install command for torch runs properly on the cluster.
    Confirms that we can properly install the package (and send a torch tensor to cuda to validate it"""
    # Use the cuda version on the cluster that we got from the previous test
    cuda_version_or_cpu = request.config.cache.get(
        ondemand_cluster.instance_type.lower(), None
    )

    assert "AttributeError" not in cuda_version_or_cpu, (
        f"No cuda version saved for {ondemand_cluster.name} - run `test_getting_cuda_version_on_clusters`"
        "to save the cuda version"
    )

    cuda_url = rh.Package.TORCH_INDEX_URLS.get(cuda_version_or_cpu)
    cuda_index_url = f"--index-url {cuda_url}"

    install_commands_for_cluster = [
        f"torch==1.13.1 {cuda_index_url} {extra_index_url}",
        f"torch>=1.13.0 {extra_index_url}",
        f"torch>1.13.0 {cuda_index_url}",
        f"torch {cuda_index_url}",
        f"torch>=1.13.0,<=2.0.0 {cuda_index_url}",
        f"torch==2.0.0 {cuda_index_url} torchaudio {cuda_index_url}",
        f"torch~=2.0.0 {cuda_index_url}",
    ]

    for install_cmd in install_commands_for_cluster:
        # Run the complete install command on the cluster
        try:
            ondemand_cluster.install_packages([install_cmd])
        except subprocess.CalledProcessError:
            assert False, f"Failed to install {install_cmd}"

    if cuda_version_or_cpu != "cpu":
        # Send a tensor to CUDA using the installed torch package (ignore if we are on CPU)
        tensor_to_cuda = rh.function(fn=send_tensor_to_cuda).to(
            ondemand_cluster, reqs=["pytest"]
        )
        sent_to_cuda = tensor_to_cuda()
        assert (
            sent_to_cuda
        ), f"Failed to send torch tensor to CUDA on {ondemand_cluster.name}"



import copy
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

from runhouse.globals import obj_store

from runhouse.logger import get_logger
from runhouse.resources.hardware.cluster import Cluster
from runhouse.resources.hardware.utils import (
    _get_cluster_from,
    detect_cuda_version_or_cpu,
)
from runhouse.resources.resource import Resource

from runhouse.utils import (
    conda_env_cmd,
    find_locally_installed_version,
    get_local_install_path,
    install_conda,
    is_python_package_string,
    locate_working_dir,
    run_setup_command,
)


INSTALL_METHODS = {"local", "reqs", "pip", "conda", "rh"}

logger = get_logger(__name__)


class CodeSyncError(Exception):
    pass


@dataclass
class InstallTarget:
    local_path: str
    _path_to_sync_to_on_cluster: Optional[str] = None

    @property
    def path_to_sync_to_on_cluster(self) -> str:
        return (
            self._path_to_sync_to_on_cluster
            if self._path_to_sync_to_on_cluster
            else f"~/{Path(self.full_local_path_str()).name}"
        )

    def full_local_path_str(self) -> str:
        return str(Path(self.local_path).expanduser().resolve())

    def __str__(self):
        return f"InstallTarget(local_path={self.local_path}, path_to_sync_to_on_cluster={self._path_to_sync_to_on_cluster})"


class Package(Resource):
    RESOURCE_TYPE = "package"

    # https://pytorch.org/get-started/locally/
    # Note: no binaries exist for 11.4 (https://github.com/pytorch/pytorch/issues/75992)
    TORCH_INDEX_URLS = {
        "11.3": "https://download.pytorch.org/whl/cu113",
        "11.5": "https://download.pytorch.org/whl/cu115",
        "11.6": "https://download.pytorch.org/whl/cu116",
        "11.7": "https://download.pytorch.org/whl/cu117",
        "11.8": "https://download.pytorch.org/whl/cu118",
        "12.1": "https://download.pytorch.org/whl/cu121",
        "cpu": "https://download.pytorch.org/whl/cpu",
    }

    def __init__(
        self,
        name: Optional[str] = None,
        install_method: Optional[str] = None,
        install_target: Optional[Union[str, "Folder"]] = None,
        install_args: Optional[str] = None,
        preferred_version: Optional[str] = None,
        dryrun: bool = False,
        **kwargs,  # We have this here to ignore extra arguments when calling from from_config
    ):
        """
        Runhouse Package resource.

        .. note::
            To create a git package, please use the factory method :func:`package`.
        """
        super().__init__(
            name=name,
            dryrun=dryrun,
        )
        self.install_method = install_method
        self.install_target = install_target
        self.install_args = install_args
        self.preferred_version = preferred_version

    def config(self, condensed: bool = True):
        # If the package is just a simple Package.from_string string, no
        # need to store it in rns, just give back the string.
        # if self.install_method in ['pip', 'conda', 'git']:
        #     return f'{self.install_method}:{self.name}'
        config = super().config(condensed)
        config["install_method"] = self.install_method
        config["install_target"] = (
            (
                self.install_target.local_path,
                self.install_target._path_to_sync_to_on_cluster,
            )
            if isinstance(self.install_target, InstallTarget)
            else self.install_target
        )
        config["install_args"] = self.install_args
        config["preferred_version"] = self.preferred_version
        return config

    def __str__(self):
        if self.name:
            return f"Package: {self.name}"
        return f"Package: {self.install_target}"

    @staticmethod
    def _prepend_python_executable(
        install_cmd: str,
        conda_env_name: Optional[str] = None,
        cluster: "Cluster" = None,
    ):
        return (
            f"python3 -m {install_cmd}"
            if cluster or conda_env_name
            else f"{sys.executable} -m {install_cmd}"
        )

    @staticmethod
    def _prepend_env_command(install_cmd: str, conda_env_name: Optional[str] = None):
        if conda_env_name:
            install_cmd = conda_env_cmd(cmd=install_cmd, conda_env_name=conda_env_name)

        return install_cmd

    def _validate_folder_path(self):
        # If self.path is the same as the user's home directory, raise an error.
        # Check this with Path and expanduser to handle both relative and absolute paths.
        if isinstance(
            self.install_target, InstallTarget
        ) and self.install_target.full_local_path_str() in [
            str(Path("~").expanduser()),
            str(Path("/")),
        ]:
            raise CodeSyncError(
                "Cannot sync the home directory. Please include a Python configuration file in a subdirectory."
            )

    def _pip_install_cmd(
        self,
        conda_env_name: Optional[str] = None,
        cluster: "Cluster" = None,
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, InstallTarget):
            if cluster:
                install_cmd = (
                    self.install_target.path_to_sync_to_on_cluster + install_args
                )
            else:
                install_cmd = self.install_target.full_local_path_str() + install_args
        else:
            install_target = f'"{self.install_target}"'
            install_cmd = install_target + install_args

        install_cmd = f"pip install {self._install_cmd_for_torch(install_cmd, cluster)}"
        install_cmd = self._prepend_python_executable(
            install_cmd, cluster=cluster, conda_env_name=conda_env_name
        )
        install_cmd = self._prepend_env_command(
            install_cmd, conda_env_name=conda_env_name
        )
        return install_cmd

    def _conda_install_cmd(
        self, conda_env_name: Optional[str] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if isinstance(self.install_target, InstallTarget):
            install_cmd = f"{self.install_target.local_path}" + install_args
        else:
            install_cmd = self.install_target + install_args

        install_cmd = f"conda install -y {install_cmd}"
        install_cmd = self._prepend_env_command(
            install_cmd, conda_env_name=conda_env_name
        )
        install_conda(cluster=cluster)
        return install_cmd

    def _reqs_install_cmd(
        self, conda_env_name: Optional[str] = None, cluster: "Cluster" = None
    ):
        install_args = f" {self.install_args}" if self.install_args else ""
        if not isinstance(self.install_target, InstallTarget):
            install_cmd = self.install_target + install_args
        else:
            on_cluster_path = self.install_target.local_path

            # If not cluster, we're on the cluster, and we must deal with the path locally
            if not cluster:
                reqs_path = f"{on_cluster_path}/requirements.txt"
                if not Path(reqs_path).expanduser().exists():
                    return None

                with open(str(Path(reqs_path).expanduser())) as f:
                    reqs_list = f.readlines()

            # Otherwise, make sure the target folder is on the cluster,
            # and read reqs from the cluster
            else:
                if "requirements.txt" not in cluster._folder_ls(
                    path=on_cluster_path, full_paths=False
                ):
                    return None

                reqs_list = (
                    cluster._folder_get(f"{on_cluster_path}/requirements.txt", mode="r")
                    .strip("\n")
                    .split("\n")
                )
                reqs_path = f"{on_cluster_path}/requirements.txt"

            install_cmd = self._reqs_install_cmd_for_torch(
                reqs_path, reqs_list, install_args, cluster=cluster
            )

        install_cmd = f"pip install {install_cmd}"
        install_cmd = self._prepend_python_executable(
            install_cmd, conda_env_name=conda_env_name, cluster=cluster
        )
        install_cmd = self._prepend_env_command(
            install_cmd, conda_env_name=conda_env_name
        )
        return install_cmd

    def _install(
        self,
        cluster: "Cluster" = None,
        node: Optional[str] = None,
        conda_env_name: Optional[str] = None,
        force_sync_local: bool = False,
    ):
        """Install package.

        Args:
            cluster (Optional[Cluster]): If provided, will install package on cluster using SSH. Otherwise, the
                assumption is that we are installing locally. (Default: ``None``)
            node (Optional[str]): Node on the cluster to install the package on, if using SSH. If ``cluster`` is
                provided without a ``node``, package will be installed on the head node. (Default: ``None``)
            conda_env_name (Optional[str]): Name of the conda environment to install the package on, if using SSH and
                installing in a specific conda env that is not activated by default.
            force_sync_local (bool, optional): If the package exists both locally and remotely, whether to override
                the remote version with the local version. By default, the local version will be installed only if
                the package does not already exist on the cluster. (Default: ``False``)
        """
        logger.info(f"Installing {str(self)} with method {self.install_method}.")

        if self.install_method == "pip":

            # If this is a generic pip package, with no version pinned, we want to check if there is a version
            # already installed. If there is, and ``force_sync_local`` is not set to ``True``, then we ignore
            # preferred version and leave the existing version.  The user can also force a version install by
            # doing `numpy==2.0.0`. Else, we install the preferred version, that matches their local.
            if (
                is_python_package_string(self.install_target)
                and self.preferred_version is not None
            ):
                # Check if this is installed
                retcode = run_setup_command(
                    f"python -c \"import importlib.util; exit(0) if importlib.util.find_spec('{self.install_target}') else exit(1)\"",
                    cluster=cluster,
                    node=node,
                )[0]
                if retcode != 0 or force_sync_local:
                    self.install_target = (
                        f"{self.install_target}=={self.preferred_version}"
                    )

            install_cmd = self._pip_install_cmd(
                conda_env_name=conda_env_name, cluster=cluster
            )
            logger.info(f"Running via install_method pip: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Pip install {install_cmd} failed, check that the package exists and is available for your platform."
                )

        elif self.install_method == "conda":
            install_cmd = self._conda_install_cmd(
                conda_env_name=conda_env_name, cluster=cluster
            )
            logger.info(f"Running via install_method conda: {install_cmd}")
            retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
            if retcode != 0:
                raise RuntimeError(
                    f"Conda install {install_cmd} failed, check that the package exists and is "
                    "available for your platform."
                )

        elif self.install_method == "reqs":
            install_cmd = self._reqs_install_cmd(
                conda_env_name=conda_env_name, cluster=cluster
            )
            if install_cmd:
                logger.info(f"Running via install_method reqs: {install_cmd}")
                retcode = run_setup_command(install_cmd, cluster=cluster, node=node)[0]
                if retcode != 0:
                    raise RuntimeError(
                        f"Reqs install {install_cmd} failed, check that the package exists and is available for your platform."
                    )
            else:
                logger.info(
                    f"{self.install_target.full_local_path_str()}/requirements.txt not found, skipping reqs install"
                )

        else:
            if self.install_method != "local":
                raise ValueError(
                    f"Unknown install method {self.install_method}. Must be one of {INSTALL_METHODS}"
                )

        # Need to append to path
        if self.install_method in ["local", "reqs"]:
            if isinstance(self.install_target, InstallTarget):
                obj_store.add_sys_path_to_all_processes(
                    self.install_target.full_local_path_str()
                ) if not cluster else run_setup_command(
                    f"export PATH=$PATH;{self.install_target.full_local_path_str()}",
                    cluster=cluster,
                    node=node,
                )
            elif not cluster:
                if Path(self.install_target).resolve().expanduser().exists():
                    obj_store.add_sys_path_to_all_processes(
                        str(Path(self.install_target).resolve().expanduser())
                    )
                else:
                    raise ValueError(
                        f"install_target {self.install_target} must be a Folder or a path to a directory for install_method {self.install_method}"
                    )
            else:
                raise ValueError(
                    f"If cluster is provided, install_target must be a Folder for install_method {self.install_method}"
                )

    # ----------------------------------
    # Torch Install Helpers
    # ----------------------------------
    def _reqs_install_cmd_for_torch(
        self, reqs_path, reqs_list, install_args="", cluster=None
    ):
        """Read requirements from file, append --index-url and --extra-index-url where relevant for torch packages,
        and return list of formatted packages."""
        # if torch extra index url is already defined by the user or torch isn't a req, directly pip install reqs file
        if not any("torch" in req for req in reqs_list):
            return f"-r {reqs_path}" + install_args

        for req in reqs_list:
            if (
                "--index-url" in req or "--extra-index-url" in req
            ) and "pytorch.org" in req:
                return f"-r {reqs_path}" + install_args

        # add extra-index-url for torch if not found
        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        return f"-r {reqs_path} --extra-index-url {self._torch_index_url(cuda_version_or_cpu)}"

    def _install_cmd_for_torch(self, install_cmd, cluster=None):
        """Return the correct formatted pip install command for the torch package(s) provided."""
        if install_cmd.startswith("#"):
            return None

        torch_source_packages = ["torch", "torchvision", "torchaudio"]
        if not any([x in install_cmd for x in torch_source_packages]):
            return install_cmd
        if "+" in install_cmd or "--extra-index-url" in install_cmd:
            return install_cmd

        packages_to_install: list = self._packages_to_install_from_cmd(install_cmd)
        final_install_cmd = ""
        cuda_version_or_cpu = detect_cuda_version_or_cpu(cluster=cluster)
        for package_install_cmd in packages_to_install:
            formatted_cmd = self._install_url_for_torch_package(
                package_install_cmd, cuda_version_or_cpu
            )
            if formatted_cmd:
                final_install_cmd += formatted_cmd + " "

        final_install_cmd = final_install_cmd.rstrip()
        return final_install_cmd if final_install_cmd != "" else None

    def _install_url_for_torch_package(self, install_cmd, cuda_version_or_cpu):
        """Build the full install command, adding a --index-url and --extra-index-url where applicable."""
        # Grab the relevant index url for torch based on the CUDA version provided
        if "," in install_cmd:
            # If installing a range of versions format the string to make it compatible with `pip_install` method
            install_cmd = install_cmd.replace(" ", "")

        index_url = self._torch_index_url(cuda_version_or_cpu)
        if index_url and not any(
            specifier in install_cmd for specifier in ["--index-url ", "-i "]
        ):
            install_cmd = f"{install_cmd} --index-url {index_url}"

        if "--extra-index-url" not in install_cmd:
            return f"{install_cmd} --extra-index-url https://pypi.python.org/simple/"

        return install_cmd

    def _torch_index_url(self, cuda_version_or_cpu: str):
        return self.TORCH_INDEX_URLS.get(cuda_version_or_cpu)

    @staticmethod
    def _packages_to_install_from_cmd(install_cmd: str):
        """Split a string of command(s) into a list of separate commands"""
        # Remove any --extra-index-url flags from the install command (to be added later by default)
        install_cmd = re.sub(r"--extra-index-url\s+\S+", "", install_cmd)
        install_cmd = install_cmd.strip()

        if ", " in install_cmd:
            # Ex: 'torch>=1.13.0,<2.0.0'
            return [install_cmd]

        matches = re.findall(r"(\S+(?:\s+(-i|--index-url)\s+\S+)?)", install_cmd)

        packages_to_install = [match[0] for match in matches]
        return packages_to_install

    def to(
        self,
        system: Union[str, Dict, "Cluster"],
        path: Optional[str] = None,
    ):
        """Copy the package onto filesystem or cluster, and return the new Package object.

        Args:
            system (str, Dict, or Cluster): Cluster to send the package to.
        """
        if not isinstance(self.install_target, InstallTarget):
            return self

        system = _get_cluster_from(system)
        if isinstance(system, Cluster) and system.on_this_cluster():
            return self

        self._validate_folder_path()
        if isinstance(system, Cluster):
            system.rsync(
                source=str(self.install_target.full_local_path_str()),
                dest=str(self.install_target.path_to_sync_to_on_cluster),
                up=True,
                contents=True,
                node="all",
            )

            new_package = copy.copy(self)
            new_package.install_target = InstallTarget(
                local_path=self.install_target.path_to_sync_to_on_cluster,
                _path_to_sync_to_on_cluster=self.install_target.path_to_sync_to_on_cluster,
            )
            return new_package

        return self

    @staticmethod
    def split_req_install_method(req_str: str):
        """Split a requirements string into a install method and the rest of the string."""
        splat = req_str.split(":", 1)
        return (splat[0], splat[1]) if len(splat) > 1 else ("", splat[0])

    @staticmethod
    def from_config(config: Dict, dryrun: bool = False, _resolve_children: bool = True):
        if isinstance(config.get("install_target"), (tuple, list)):
            config["install_target"] = InstallTarget(
                local_path=config["install_target"][0],
                _path_to_sync_to_on_cluster=config["install_target"][1],
            )
        if config.get("resource_subtype") == "GitPackage":
            from runhouse import GitPackage

            return GitPackage.from_config(
                config, dryrun=dryrun, _resolve_children=_resolve_children
            )
        return Package(**config, dryrun=dryrun)

    @staticmethod
    def from_string(specifier: str, dryrun: bool = False):
        if specifier == "requirements.txt":
            specifier = "reqs:./"

        # Use regex to check if specifier matches '<method>:https://github.com/<path>' or 'https://github.com/<path>'
        match = re.search(
            r"^(?:(?P<method>[^:]+):)?(?P<path>https://github.com/.+)", specifier
        )
        if match:
            install_method = match.group("method")
            url = match.group("path")
            from runhouse.resources.packages.git_package import git_package

            return git_package(
                git_url=url, install_method=install_method, dryrun=dryrun
            )

        install_method, target_and_args = Package.split_req_install_method(specifier)

        # Handles a case like "torch --index-url https://download.pytorch.org/whl/cu113"
        target, args = (
            target_and_args.split(" ", 1)
            if " " in target_and_args
            else (target_and_args, "")
        )

        # If the target is a path
        # If it doesn't have a /, we're assuming it's a pip installable thing first and foremost
        if "/" in target or "~" in target:
            # We need to do this because relative paths are relative to the current working directory!
            abs_target = (
                Path(target).expanduser()
                if Path(target).expanduser().is_absolute()
                else Path(locate_working_dir()) / target
            )
            if abs_target.exists():
                target = InstallTarget(
                    local_path=str(abs_target), _path_to_sync_to_on_cluster=None
                )

        # If install method is not provided, we need to infer it
        if not install_method:
            if Path(specifier).expanduser().resolve().exists():
                install_method = "reqs"
            else:
                install_method = "pip"

        # If we are just defaulting to pip, attempt to install the same version of the package
        # that is already installed locally
        # Check if the target is only letters, nothing else. This means its a string like 'numpy'.
        preferred_version = None
        if install_method == "pip" and is_python_package_string(target):
            locally_installed_version = find_locally_installed_version(target)
            if locally_installed_version:
                # Check if this is a package that was installed from local
                local_install_path = get_local_install_path(target)
                if local_install_path and Path(local_install_path).exists():
                    target = InstallTarget(
                        local_path=local_install_path, _path_to_sync_to_on_cluster=None
                    )

                else:
                    # We want to preferrably install this version of the package server-side
                    preferred_version = locally_installed_version

        # "Local" install method is a special case where we just copy a local folder and add to path
        if install_method == "local":
            return Package(
                install_target=target, install_method=install_method, dryrun=dryrun
            )

        elif install_method in ["reqs", "pip", "conda"]:
            return Package(
                install_target=target,
                install_args=args,
                install_method=install_method,
                preferred_version=preferred_version,
                dryrun=dryrun,
            )
        elif install_method == "rh":
            # Calling the factory method below
            return package(name=specifier[len("rh:") :], dryrun=dryrun)
        else:
            raise ValueError(
                f"Unknown install method {install_method}. Must be one of {INSTALL_METHODS}"
            )


def package(
    name: str = None,
    install_method: str = None,
    install_str: str = None,
    path: str = None,
    system: str = None,
    load_from_den: bool = True,
    dryrun: bool = False,
) -> Package:
    """
    Builds an instance of :class:`Package`.

    Args:
        name (str, optional): Name to assign the package resource.
        install_method (str, optional): Method for installing the package.
            Options: [``pip``, ``conda``, ``reqs``, ``local``]
        install_str (str, optional): Additional arguments to install.
        path (str, optional): URL of the package to install.
        system (str, optional): File system or cluster on which the package lives.
            Currently this must a cluster or one of: [``file``, ``s3``, ``gs``].
        load_from_den (bool, optional): Whether to try loading the Package from Den. (Default: ``True``)
        dryrun (bool, optional): Whether to create the Package if it doesn't exist, or load the Package
            object as a dryrun. (Default: ``False``)

    Returns:
        Package: The resulting package.

    Example:
        >>> reloaded_package = rh.package(name="my-package")
        >>> local_package = rh.package(path="local/folder/path", install_method="local")
    """
    if name and not any([install_method, install_str, path, system]):
        # If only the name is provided and dryrun is set to True
        return Package.from_name(name, load_from_den=load_from_den, dryrun=dryrun)

    install_target = None
    install_args = None
    if path is not None:
        install_target = (path, None)
        install_args = install_str
    elif install_str is not None:
        install_target, install_args = install_str.split(" ", 1)

    return Package(
        install_method=install_method,
        install_target=install_target,
        install_args=install_args,
        name=name,
        dryrun=dryrun,
    )



import shlex
import subprocess

from runhouse.constants import SKY_VENV
from runhouse.logger import get_logger

logger = get_logger(__name__)


class AutostopHelper:
    """A helper class strictly to run SkyPilot methods on OnDemandClusters inside SkyPilot's conda env."""

    def __init__(self):
        self._activity_registered = False

    async def set_last_active_time_to_now(self):
        self._activity_registered = True

    def _run_python_in_sky_venv(self, cmd: str):
        sky_python_cmd = f"{SKY_VENV}/bin/python -c {cmd}"

        logger.debug(f"Running command in SkyPilot's venv: {sky_python_cmd}")
        # run with subprocess and return the output
        return subprocess.run(
            sky_python_cmd,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).stdout.decode("utf-8")

    async def get_autostop(self):
        sky_get_autostop_cmd = shlex.quote(
            "from sky.skylet.autostop_lib import get_autostop_config; "
            "print(get_autostop_config().autostop_idle_minutes)"
        )

        return int(self._run_python_in_sky_venv(sky_get_autostop_cmd))

    async def get_last_active_time(self):
        sky_get_last_active_time_cmd = shlex.quote(
            "from sky.skylet.autostop_lib import get_last_active_time; "
            "print(get_last_active_time())"
        )

        return float(self._run_python_in_sky_venv(sky_get_last_active_time_cmd))

    async def set_autostop(self, idle_minutes: int):
        # Filling in "cloudvmray" as the backend because it's the only backend supported by SkyPilot right now,
        # if needed we can grab the backend from the autostop config with:
        # `from sky.skylet.autostop_lib import get_autostop_config; get_autostop_config().backend`
        sky_set_autostop_cmd = shlex.quote(
            f"from sky.skylet.autostop_lib import set_autostop; "
            f'set_autostop({idle_minutes}, "cloudvmray", True)'
        )

        self._run_python_in_sky_venv(sky_set_autostop_cmd)

    async def register_activity_if_needed(self):
        sky_register_activity_cmd = shlex.quote(
            "from sky.skylet.autostop_lib import set_last_active_time_to_now; "
            "set_last_active_time_to_now()"
        )

        if self._activity_registered:
            logger.debug("Activity registered, updating last active time in SkyConfig")
            self._run_python_in_sky_venv(sky_register_activity_cmd)
            self._activity_registered = False
        else:
            logger.debug(
                "No activity registered, not updating last active time in SkyConfig"
            )



import re
from pathlib import Path

from runhouse.globals import configs, rns_client
from runhouse.resources.hardware.utils import _get_cluster_from
from runhouse.utils import generate_default_name

DEFAULT_LOCAL_FOLDER = f"{Path.cwd()}/"
DEFAULT_CLUSTER_FS_FOLDER = (
    ""  # Objects will land inside home directory when sent without a path
)
DEFAULT_BLOB_STORAGE_FOLDER = (
    configs.get("default_blob_storage_folder", "runhouse") + "/"
)


def is_valid_resource_name(name, strict_slashes=False):
    if strict_slashes is True:
        # Require a leading slash if any are present in string
        return re.match(
            r"^(?!.*\/{2,})(?![^\/\.\@~]+\/)[.@~]?[a-zA-Z0-9\-\_\/@]{2,200}[a-zA-Z0-9\-\_]+$",
            name,
        )
    return re.match(
        r"^(?!.*\/{2,})[.@~]?[a-zA-Z0-9\-\_\/@]{2,200}[a-zA-Z0-9\-\_]+$",
        name,
    )


def _generate_default_path(cls, name, system):
    """Generate a default path for a data resource. Logic is as follows:
    1. If the system is a local file system, save to the current working directory
    2. If the system is a remote file system, save to the default cache folder
    3. If the system is a remote object store, save to the default object store folder
    """

    from runhouse.resources.hardware import Cluster

    system = _get_cluster_from(system)

    name = name or generate_default_name(prefix=cls.RESOURCE_TYPE)
    if system == rns_client.DEFAULT_FS or "here":
        base_folder = DEFAULT_LOCAL_FOLDER
    elif isinstance(system, Cluster):
        if system.on_this_cluster():
            base_folder = DEFAULT_LOCAL_FOLDER
        else:
            base_folder = DEFAULT_CLUSTER_FS_FOLDER
    else:
        base_folder = DEFAULT_BLOB_STORAGE_FOLDER
    return f"{base_folder}{cls.RESOURCE_TYPE}/{name}"



"""
Runhouse breaks down the silos between machine learning compute and data environments, and makes interacting
with cloud resources easy and Pythonic:
With Runhouse, you can:
- Run effortlessly on any system in experimentation, research, or production, iterating and debugging like
  you’re running locally on bare metal.
- Scale up, productionize, or share your work without weeks of packaging or translating into platform DSLs.
- Retain control of their underlying infra, customizing it to your unique needs and continuing to use any ML
  tools you would from cloud instances.
"""
import datetime

# Adapted from: https://github.com/skypilot-org/skypilot/blob/master/sky/setup_files/setup.py

import io
import os
import platform
import re
import warnings

import setuptools

ROOT_DIR = os.path.dirname(__file__)

system = platform.system()
if system == "Darwin":
    mac_version = platform.mac_ver()[0]
    mac_major, mac_minor = mac_version.split(".")[:2]
    mac_major = int(mac_major)
    mac_minor = int(mac_minor)
    if mac_major < 10 or (mac_major == 10 and mac_minor < 15):
        warnings.warn(
            f"'Detected MacOS version {mac_version}. MacOS version >=10.15 "
            "is required to install ray>=1.9'"
        )


def find_version(*filepath):
    # Extract version information from filepath
    with open(os.path.join(ROOT_DIR, *filepath)) as fp:
        version_match = re.search(
            r'^__version__ = [\'"]([^\'"]*)[\'"]', fp.read(), re.M
        )
        if version_match:
            return version_match.group(1)
        raise RuntimeError("Unable to find version string.")


def parse_readme(readme: str) -> str:
    """Parse the README.md file to be pypi compatible."""
    # Replace the footnotes.
    readme = readme.replace("<!-- Footnote -->", "#")
    footnote_re = re.compile(r"\[\^([0-9]+)\]")
    readme = footnote_re.sub(r"<sup>[\1]</sup>", readme)

    # Remove the dark mode switcher
    mode_re = re.compile(
        r"<picture>[\n ]*<source media=.*>[\n ]*<img(.*)>[\n ]*</picture>", re.MULTILINE
    )
    readme = mode_re.sub(r"<img\1>", readme)
    return readme


install_requires = [
    "annotated-types>=0.6.0",
    "python-dotenv",
    "pexpect",
    "pyOpenSSL>=23.3.0",
    "async_timeout",  # Needed for ray<=2.9
    "rich",
    "setuptools < 70.0.0",  # Bug in setuptools 70.0.0: https://github.com/AUTOMATIC1111/stable-diffusion-webui/issues/15863 noqa
    "typer",
    "wheel",
    "apispec",
    "pyyaml",
    "cloudpickle",
    "httpx",
    "pydantic >= 2.5.0",  # required for ray >= 2.9.0 (https://github.com/ray-project/ray/releases?page=2)
    "cryptography",
    "requests",
]

# NOTE: Change the templates/spot-controller.yaml.j2 file if any of the following
# packages dependencies are changed.
extras_require = {
    "sky": ["skypilot==0.7.0"],
    "data": ["pandas"],
    "server": ["ray[default] >= 2.9.0", "uvicorn", "fastapi", "pynvml", "psutil"],
    "aws": [
        "skypilot[aws]==0.7.0",
        # https://github.com/aio-libs/aiobotocore/issues/983
        # If you don't want to use these exact versions, you can install runhouse without the aws extras, then
        # install your desired versions of awscli and boto3
        "pycryptodome==3.12.0",
    ],
    "azure": ["skypilot[azure]==0.7.0"],
    "gcp": [
        "skypilot[gcp]==0.7.0",
        "gcsfs",
    ],
    "docker": ["docker"],
    "kubernetes": ["skypilot==0.7.0", "kubernetes"],
}

extras_require["all"] = sum(extras_require.values(), [])

long_description = ""
readme_filepath = "README.md"
if os.path.exists(readme_filepath):
    long_description = io.open(readme_filepath, "r", encoding="utf-8").read()
    long_description = parse_readme(long_description)

# Flip to True to build the nightly instead of release.
nightly = False
version = find_version("runhouse", "__init__.py")

setuptools.setup(
    # NOTE: this affects the package.whl wheel name. When changing this (if
    # ever), you must grep for '.whl' and change all corresponding wheel paths
    # (templates/*.j2 and wheel_utils.py).
    name="runhouse" if not nightly else "runhouse-nightly",
    # append .dev and the date in YYYYMMDD format for nightly builds
    version=version + f'.dev{datetime.datetime.today().strftime("%Y%m%d")}'
    if nightly
    else version,
    packages=setuptools.find_packages(exclude=["tests"]),
    author="Runhouse Team",
    license="Apache 2.0",
    readme="README.md",
    description="Runhouse: A multiplayer cloud compute and data environment",
    long_description=long_description,
    long_description_content_type="text/markdown",
    setup_requires=["wheel"],
    requires_python=">=3.7",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": [
            "runhouse = runhouse.main:app",
            "rh = runhouse.main:app",
        ],
    },
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System :: Distributed Computing",
    ],
    project_urls={
        "Homepage": "https://run.house",
        "Issues": "https://github.com/run-house/runhouse/issues/",
    },
)



from .cluster import Cluster
from .cluster_factory import cluster, ondemand_cluster
from .on_demand_cluster import OnDemandCluster
from .ray_utils import check_for_existing_ray_instance, kill_actors, list_actor_states
from .utils import (
    _current_cluster,
    _get_cluster_from,
    cluster_config_file_exists,
    ClusterStatus,
    get_all_sky_clusters,
    load_cluster_config_from_file,
    SSEClient,
)
